import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import cm
from scipy.interpolate import interp1d
from scipy.spatial import distance
from scipy.optimize import differential_evolution


class IntracellAnalysis:
    # IA constants
    UPPER_VOLTAGE = 4.2
    LOWER_VOLTAGE = 2.7
    THRESHOLD = 4.84 * 0.7

    def __init__(self,
                 pe_pristine_file,
                 ne_pristine_file,
                 cycle_type='rpt_0.2C',
                 step_type=0,
                 ne_2pos_file=None,
                 ne_2neg_file=None
                 ):
        """
        Invokes the cell electrode analysis class. This is a class designed to fit the cell and electrode
        parameters in order to determine changes of electrodes within the full cell from only full cell cycling data.
        Args:
            pe_pristine_file (str): file name for the half cell data of the pristine (uncycled) positive
                electrode
            ne_pristine_file (str): file name for the half cell data of the pristine (uncycled) negative
                electrode
            cycle_type (str): type of diagnostic cycle for the fitting
            step_type (int): charge or discharge (0 for charge, 1 for discharge)
            ne_2neg_file (str): file name of the data for the negative component of the anode
            ne_2pos_file (str): file name of the data for the positive component of the anode
        """
        self.pe_pristine = pd.read_csv(pe_pristine_file, usecols=['SOC_aligned', 'Voltage_aligned'])
        self.ne_1_pristine = pd.read_csv(ne_pristine_file, usecols=['SOC_aligned', 'Voltage_aligned'])

        if ne_2neg_file and ne_2pos_file:
            self.ne_2_pristine_pos = pd.read_csv(ne_2pos_file)
            self.ne_2_pristine_neg = pd.read_csv(ne_2neg_file)
        else:
            self.ne_2_pristine_pos = pd.DataFrame()
            self.ne_2_pristine_neg = pd.DataFrame()

        self.cycle_type = cycle_type
        self.step_type = step_type

    def process_beep_cycle_data_for_candidate_halfcell_analysis(self,
                                                                cell_struct,
                                                                real_cell_initial_charge_profile_aligned,
                                                                real_cell_initial_charge_profile,
                                                                cycle_index):
        """
        Inputs:
        diag_type_cycles: beep cell_struct.diagnostic_interpolated filtered to one diagnostic type
        real_cell_initial_charge_profile_aligned: dataframe containing SOC (equally spaced) and voltage columns
        real_cell_initial_charge_profile: dataframe containing SOC and voltage columns
        cycle_index: cycle number to evaluate at
        Outputs
        real_cell_candidate_charge_profile_aligned: a dataframe containing columns SOC_aligned (evenly spaced) and
        Voltage_aligned
        """

        diag_type_cycles = cell_struct.diagnostic_data.loc[cell_struct.diagnostic_data['cycle_type'] == self.cycle_type]
        real_cell_candidate_charge_profile = diag_type_cycles.loc[
            (diag_type_cycles.cycle_index == cycle_index)
            & (diag_type_cycles.step_type == 0)  # step_type = 0 is charge, 1 is discharge
            & (diag_type_cycles.voltage < self.UPPER_VOLTAGE)
            & (diag_type_cycles[self.capacity_col] > 0)][['voltage', 'charge_capacity']]

        real_cell_candidate_charge_profile['SOC'] = (
                (real_cell_candidate_charge_profile['charge_capacity'] -
                 np.min(real_cell_initial_charge_profile['charge_capacity']))
                / (np.max(real_cell_initial_charge_profile['charge_capacity']) -
                   np.min(real_cell_initial_charge_profile['charge_capacity'])) * 100
                                                     )
        real_cell_candidate_charge_profile['Voltage'] = real_cell_candidate_charge_profile['voltage']
        real_cell_candidate_charge_profile.drop('voltage', axis=1, inplace=True)

        SOC_vec = np.linspace(0, np.max(real_cell_candidate_charge_profile['SOC']),
                              1001)  # 100 ; np.max(real_cell_candidate_charge_profile['SOC']

        real_cell_candidate_charge_profile_aligned = pd.DataFrame()
        real_cell_candidate_charge_profile_interper = interp1d(real_cell_candidate_charge_profile['SOC'],
                                                               real_cell_candidate_charge_profile['Voltage'],
                                                               bounds_error=False,
                                                               fill_value=(self.LOWER_VOLTAGE, self.UPPER_VOLTAGE))
        real_cell_candidate_charge_profile_aligned['Voltage_aligned'] = real_cell_candidate_charge_profile_interper(
            SOC_vec)

        # real_cell_candidate_charge_profile_aligned['Voltage_aligned'].fillna(self.LOWER_VOLTAGE, inplace=True)
        real_cell_candidate_charge_profile_aligned['SOC_aligned'] = SOC_vec / np.max(
            real_cell_initial_charge_profile_aligned['SOC_aligned'].loc[
                ~real_cell_initial_charge_profile_aligned['Voltage_aligned'].isna()]) * 100

        return real_cell_candidate_charge_profile_aligned

    def process_beep_cycle_data_for_initial_halfcell_analysis(self,
                                                              cell_struct,
                                                              step_type=0):
        """
        This function extracts the initial (non-degraded) voltage and soc profiles for the cell with columns
        interpolated on voltage and soc.
        Inputs
        cell_struct: beep cell_struct.diagnostic_interpolated filtered to one diagnostic type
        step_type: specifies whether the cell is charging or discharging. 0 is charge, 1 is discharge.
        Outputs
        real_cell_initial_charge_profile_aligned: a dataframe containing columns SOC_aligned (evenly spaced)
            and Voltage_aligned
        real_cell_initial_charge_profile: a dataframe containing columns Voltage (evenly spaced), capacity, and SOC
        """
        if step_type == 0:
            self.capacity_col = 'charge_capacity'
        else:
            self.capacity_col = 'discharge_capacity'

        diag_type_cycles = cell_struct.diagnostic_data.loc[cell_struct.diagnostic_data['cycle_type'] == self.cycle_type]
        soc_vec = np.linspace(0, 100.0, 1001)
        cycle_index_of_cycle_type = cell_struct.diagnostic_summary[
            cell_struct.diagnostic_summary.cycle_type == self.cycle_type].cycle_index.iloc[0]
        real_cell_initial_charge_profile = diag_type_cycles.loc[
            (diag_type_cycles.cycle_index == cycle_index_of_cycle_type)
            & (diag_type_cycles.step_type == step_type)  # step_type = 0 is charge, 1 is discharge
            & (diag_type_cycles.voltage < self.UPPER_VOLTAGE)
            & (diag_type_cycles.voltage > self.LOWER_VOLTAGE)][['voltage', self.capacity_col]]

        real_cell_initial_charge_profile['SOC'] = (
                (
                        real_cell_initial_charge_profile[self.capacity_col] -
                        np.min(real_cell_initial_charge_profile[self.capacity_col])
                ) /
                (
                       np.max(real_cell_initial_charge_profile[self.capacity_col]) -
                       np.min(real_cell_initial_charge_profile[self.capacity_col])
                ) * 100
                                                   )
        real_cell_initial_charge_profile['Voltage'] = real_cell_initial_charge_profile['voltage']
        real_cell_initial_charge_profile.drop('voltage', axis=1, inplace=True)

        real_cell_initial_charge_profile_aligned = pd.DataFrame()
        real_cell_initial_charge_profile_aligned['SOC_aligned'] = soc_vec
        real_cell_initial_charge_profile_interper = interp1d(real_cell_initial_charge_profile['SOC'],
                                                             real_cell_initial_charge_profile['Voltage'],
                                                             bounds_error=False)

        real_cell_initial_charge_profile_aligned['Voltage_aligned'] = real_cell_initial_charge_profile_interper(
            real_cell_initial_charge_profile_aligned['SOC_aligned'])

        return real_cell_initial_charge_profile_aligned, real_cell_initial_charge_profile

    def get_dQdV_over_Q_from_halfcell_initial_matching(self, x, *params):
        df_1, df_2, df_real_interped, emulated_full_cell_interped = self.halfcell_initial_matching_v2(x, *params)

        # Calculate dQdV from full cell profiles
        dq_dv_real = pd.DataFrame(np.gradient(df_real_interped['SOC_aligned'], df_real_interped['Voltage_aligned']),
                                  columns=['dQdV'])
        dq_dv_emulated = pd.DataFrame(
            np.gradient(emulated_full_cell_interped['SOC_aligned'], emulated_full_cell_interped['Voltage_aligned']),
            columns=['dQdV'])

        # Include original data
        dq_dv_real['SOC_aligned'] = df_real_interped['SOC_aligned']
        dq_dv_real['Voltage_aligned'] = df_real_interped['Voltage_aligned']

        dq_dv_emulated['SOC_aligned'] = emulated_full_cell_interped['SOC_aligned']
        dq_dv_emulated['Voltage_aligned'] = emulated_full_cell_interped['Voltage_aligned']

        # Interpolate over Q
        # ^^ already done in this case as standard data template is over Q

        return df_1, df_2, dq_dv_real, dq_dv_emulated, df_real_interped, emulated_full_cell_interped

    def get_error_dQdV_over_Q_from_halfcell_initial_matching(self, x, *params):
        df_1, df_2, dQdV_real, dQdV_emulated, df_real_interped, emulated_full_cell_interped = \
            self.get_dQdV_over_Q_from_halfcell_initial_matching(x, *params)

        # Calculate distance between lines
        error = distance.euclidean(dQdV_real['dQdV'], dQdV_emulated['dQdV']) + 0.01 * len(
            dQdV_emulated['dQdV'].loc[dQdV_emulated['dQdV'].isna()])

        return error

    def get_dQdV_over_V_from_halfcell_initial_matching(self, x, *params):
        df_1, df_2, df_real_interped, emulated_full_cell_interped = self.halfcell_initial_matching_v2(x, *params)

        # Calculate dQdV from full cell profiles
        dq_dv_real = pd.DataFrame(np.gradient(df_real_interped['SOC_aligned'], df_real_interped['Voltage_aligned']),
                                  columns=['dQdV']).ewm(alpha=x[-2]).mean()
        dq_dv_emulated = pd.DataFrame(
            np.gradient(emulated_full_cell_interped['SOC_aligned'], emulated_full_cell_interped['Voltage_aligned']),
            columns=['dQdV']).ewm(alpha=x[-1]).mean()

        # Include original data
        dq_dv_real['SOC_aligned'] = df_real_interped['SOC_aligned']
        dq_dv_real['Voltage_aligned'] = df_real_interped['Voltage_aligned']

        dq_dv_emulated['SOC_aligned'] = emulated_full_cell_interped['SOC_aligned']
        dq_dv_emulated['Voltage_aligned'] = emulated_full_cell_interped['Voltage_aligned']

        # Interpolate over V
        voltage_vec = np.linspace(2.7, 4.2, 1001)

        v_dq_dv_interper_real = interp1d(dq_dv_real['Voltage_aligned'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                         dq_dv_real['dQdV'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                         bounds_error=False, fill_value=0)
        v_soc_interper_real = interp1d(dq_dv_real['Voltage_aligned'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                       dq_dv_real['SOC_aligned'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                       bounds_error=False, fill_value=(0, 100))

        v_dq_dv_interper_emulated = interp1d(dq_dv_emulated['Voltage_aligned'].loc[
                                                ~dq_dv_emulated['Voltage_aligned'].isna()],
                                             dq_dv_emulated['dQdV'].loc[~dq_dv_emulated['Voltage_aligned'].isna()],
                                             bounds_error=False, fill_value=0)
        v_soc_interper_emulated = interp1d(dq_dv_emulated['Voltage_aligned'].loc[
                                               ~dq_dv_emulated['Voltage_aligned'].isna()],
                                           dq_dv_emulated['SOC_aligned'].loc[~dq_dv_emulated['Voltage_aligned'].isna()],
                                           bounds_error=False, fill_value=(0, 100))

        dq_dv_over_v_real = pd.DataFrame(v_dq_dv_interper_real(voltage_vec), columns=['dQdV']).fillna(0)
        dq_dv_over_v_real['SOC'] = v_soc_interper_real(voltage_vec)
        dq_dv_over_v_real['Voltage'] = voltage_vec

        dq_dv_over_v_emulated = pd.DataFrame(v_dq_dv_interper_emulated(voltage_vec), columns=['dQdV']).fillna(0)
        dq_dv_over_v_emulated['SOC'] = v_soc_interper_emulated(voltage_vec)
        dq_dv_over_v_emulated['Voltage'] = voltage_vec

        return df_1, df_2, dq_dv_over_v_real, dq_dv_over_v_emulated, df_real_interped, emulated_full_cell_interped

    def get_error_dQdV_over_V_from_halfcell_initial_matching(self, x, *params):
        df_1, df_2, dq_dv_real, dq_dv_emulated, df_real_interped, emulated_full_cell_interped = \
            self.get_dQdV_over_V_from_halfcell_initial_matching(x, *params)

        # Calculate distance between lines
        error = distance.euclidean(dq_dv_real['dQdV'], dq_dv_emulated['dQdV']) + 0.01 * len(
            dq_dv_emulated['dQdV'].loc[dq_dv_emulated['dQdV'].isna()])
        return error

    def get_dVdQ_over_Q_from_halfcell_initial_matching(self, x, *params):
        df_1, df_2, df_real_interped, emulated_full_cell_interped = self.halfcell_initial_matching_v2(x, *params)

        # Calculate dVdQ from full cell profiles
        dv_dq_real = pd.DataFrame(np.gradient(df_real_interped['Voltage_aligned'], df_real_interped['SOC_aligned']),
                                  columns=['dVdQ'])
        dv_dq_emulated = pd.DataFrame(
            np.gradient(emulated_full_cell_interped['Voltage_aligned'], emulated_full_cell_interped['SOC_aligned']),
            columns=['dVdQ'])

        # Include original data
        dv_dq_real['SOC_aligned'] = df_real_interped['SOC_aligned']
        dv_dq_real['Voltage_aligned'] = df_real_interped['Voltage_aligned']

        dv_dq_emulated['SOC_aligned'] = emulated_full_cell_interped['SOC_aligned']
        dv_dq_emulated['Voltage_aligned'] = emulated_full_cell_interped['Voltage_aligned']

        # Interpolate over Q
        # ^^ already done in this case as standard data template is over Q

        return df_1, df_2, dv_dq_real, dv_dq_emulated, df_real_interped, emulated_full_cell_interped

    def get_error_dVdQ_over_Q_from_halfcell_initial_matching(self, x, *params):
        (df_1,
         df_2,
         dVdQ_real,
         dVdQ_emulated,
         df_real_interped,
         emulated_full_cell_interped) = self.get_dVdQ_over_Q_from_halfcell_initial_matching(x, *params)

        # Calculate distance between lines
        error = distance.euclidean(dVdQ_real['dVdQ'], dVdQ_emulated['dVdQ']) + 0.01 * len(
            dVdQ_emulated['dVdQ'].loc[dVdQ_emulated['dVdQ'].isna()])

        return error

    def get_dVdQ_over_V_from_halfcell_initial_matching(self, x, *params):
        (df_1,
         df_2,
         df_real_interped,
         emulated_full_cell_interped) = self.halfcell_initial_matching_v2(x, *params)

        # Calculate dVdQ from full cell profiles
        dv_dq_real = pd.DataFrame(np.gradient(df_real_interped['SOC_aligned'], df_real_interped['Voltage_aligned']),
                                  columns=['dVdQ'])
        dv_dq_emulated = pd.DataFrame(
            np.gradient(emulated_full_cell_interped['SOC_aligned'], emulated_full_cell_interped['Voltage_aligned']),
            columns=['dVdQ'])

        # Include original data
        dv_dq_real['SOC_aligned'] = df_real_interped['SOC_aligned']
        dv_dq_real['Voltage_aligned'] = df_real_interped['Voltage_aligned']

        dv_dq_emulated['SOC_aligned'] = emulated_full_cell_interped['SOC_aligned']
        dv_dq_emulated['Voltage_aligned'] = emulated_full_cell_interped['Voltage_aligned']

        # Interpolate over V
        voltage_vec = np.linspace(2.7, 4.2, 1001)
        v_dv_dq_interper_real = interp1d(dv_dq_real['Voltage_aligned'].loc[~dv_dq_real['Voltage_aligned'].isna()],
                                         dv_dq_real['dVdQ'].loc[~dv_dq_real['Voltage_aligned'].isna()],
                                         bounds_error=False, fill_value=0)
        v_soc_interper_real = interp1d(dv_dq_real['Voltage_aligned'].loc[~dv_dq_real['Voltage_aligned'].isna()],
                                       dv_dq_real['SOC_aligned'].loc[~dv_dq_real['Voltage_aligned'].isna()],
                                       bounds_error=False, fill_value=(0, 100))

        v_dv_dq_interp_emulated = interp1d(dv_dq_emulated['Voltage_aligned'].loc[
                                              ~dv_dq_emulated['Voltage_aligned'].isna()],
                                           dv_dq_emulated['dVdQ'].loc[~dv_dq_emulated['Voltage_aligned'].isna()],
                                           bounds_error=False, fill_value=0)
        v_soc_interper_emulated = interp1d(dv_dq_emulated['Voltage_aligned'].loc[
                                               ~dv_dq_emulated['Voltage_aligned'].isna()],
                                           dv_dq_emulated['SOC_aligned'].loc[~dv_dq_emulated['Voltage_aligned'].isna()],
                                           bounds_error=False, fill_value=(0, 100))

        dv_dq_over_v_real = pd.DataFrame(v_dv_dq_interper_real(voltage_vec), columns=['dVdQ'])
        dv_dq_over_v_real['SOC'] = v_soc_interper_real(voltage_vec)
        dv_dq_over_v_real['Voltage'] = voltage_vec

        dv_dq_over_v_emulated = pd.DataFrame(v_dv_dq_interp_emulated(voltage_vec), columns=['dVdQ'])
        dv_dq_over_v_emulated['SOC'] = v_soc_interper_emulated(voltage_vec)
        dv_dq_over_v_emulated['Voltage'] = voltage_vec

        return df_1, df_2, dv_dq_over_v_real, dv_dq_over_v_emulated, df_real_interped, emulated_full_cell_interped

    def get_error_dVdQ_over_V_from_halfcell_initial_matching(self, x, *params):
        df_1, df_2, dv_dq_real, dv_dq_emulated, df_real_interped, emulated_full_cell_interped = \
            self.get_dVdQ_over_V_from_halfcell_initial_matching(x, *params)

        # Calculate distance between lines
        error = distance.euclidean(dv_dq_real['dVdQ'], dv_dq_emulated['dVdQ']) + 0.01 * len(
            dv_dq_emulated['dVdQ'].loc[dv_dq_emulated['dVdQ'].isna()])
        return error

    def blend_electrodes_robust(self, ne_pristine_matched, ne_2_pos, ne_2_neg, x_ne_2):
        """
        """
        if ne_2_pos.empty:
            df_blended = ne_pristine_matched
            return df_blended

        if ne_2_neg.empty:
            df_blended = ne_pristine_matched
            return df_blended

        if x_ne_2 > 0:
            ne_2_pristine = ne_2_pos
        else:
            ne_2_pristine = ne_2_neg
            x_ne_2 = np.abs(x_ne_2)

        # match the two NE materials by SOC
        ne_pristine_matched_interper = interp1d(
            ne_pristine_matched.loc[~ne_pristine_matched.Voltage_aligned.isna()]['SOC_aligned'],
            ne_pristine_matched.loc[~ne_pristine_matched.Voltage_aligned.isna()]['Voltage_aligned'],
            bounds_error=False)
        ne_2_pristine_interper = interp1d(ne_2_pristine.loc[~ne_2_pristine.Voltage_aligned.isna()]['SOC_aligned'],
                                          ne_2_pristine.loc[~ne_2_pristine.Voltage_aligned.isna()]['Voltage_aligned'],
                                          bounds_error=False)

        soc_vec = np.linspace(0, 100, 1001)

        ne_pristine_matched_len1001 = pd.DataFrame(soc_vec, columns=['SOC_aligned'])
        ne_pristine_matched_len1001['Voltage_aligned'] = ne_pristine_matched_interper(soc_vec)

        ne_2_pristine_interper_len1001 = pd.DataFrame(soc_vec, columns=['SOC_aligned'])
        ne_2_pristine_interper_len1001['Voltage_aligned'] = ne_2_pristine_interper(soc_vec)

        df_ne = blend_electrodes(ne_pristine_matched_len1001,
                                 ne_2_pristine_interper_len1001,
                                 pd.DataFrame(),
                                 x_ne_2)

        return df_ne

    def blend_electrodes_robust_v2(self, ne_pristine_matched, ne_2_pos, ne_2_neg, x_ne_2):
        """
        """
        if ne_2_pos.empty:
            df_blended = ne_pristine_matched
            return df_blended

        if ne_2_neg.empty:
            df_blended = ne_pristine_matched
            return df_blended

        if x_ne_2 > 0:
            ne_2_pristine = ne_2_pos
        else:
            ne_2_pristine = ne_2_neg
            x_ne_2 = np.abs(x_ne_2)

        # match the two NE materials by SOC
        soc_vec = np.linspace(0, 100, 1001)

        ne_pristine_matched_0to100 = ((ne_pristine_matched.loc[~ne_pristine_matched.Voltage_aligned.isna()][
                                           'SOC_aligned'] - np.min(
            ne_pristine_matched.loc[~ne_pristine_matched.Voltage_aligned.isna()]['SOC_aligned'])) /
                                      (np.max(ne_pristine_matched.loc[~ne_pristine_matched.Voltage_aligned.isna()][
                                                  'SOC_aligned']) - np.min(
                                          ne_pristine_matched.loc[~ne_pristine_matched.Voltage_aligned.isna()][
                                              'SOC_aligned']))
                                      ) * 100
        ne_pristine_matched_interper = interp1d(ne_pristine_matched_0to100,
                                                ne_pristine_matched.loc[~ne_pristine_matched.Voltage_aligned.isna()][
                                                    'Voltage_aligned'],
                                                bounds_error=False)

        ne_2_pristine_interper = interp1d(ne_2_pristine.loc[~ne_2_pristine.Voltage_aligned.isna()]['SOC_aligned'],
                                          ne_2_pristine.loc[~ne_2_pristine.Voltage_aligned.isna()]['Voltage_aligned'],
                                          bounds_error=False)

        ne_pristine_matched_len1001 = pd.DataFrame(soc_vec, columns=['SOC_aligned'])
        ne_pristine_matched_len1001['Voltage_aligned'] = ne_pristine_matched_interper(soc_vec)

        ne_2_pristine_interper_len1001 = pd.DataFrame(soc_vec, columns=['SOC_aligned'])
        ne_2_pristine_interper_len1001['Voltage_aligned'] = ne_2_pristine_interper(soc_vec)

        df_ne_blended = blend_electrodes(ne_pristine_matched_len1001,
                                         ne_2_pristine_interper_len1001,
                                         pd.DataFrame(),
                                         x_ne_2)

        # restore blend back to original SOC span of NE_pristine_matched

        df_ne_blended_interper = interp1d(df_ne_blended['SOC_aligned'], df_ne_blended['Voltage_aligned'],
                                          bounds_error=False)  # intiializing interpolation across the blend

        len_non_na_ne_pristine_matched = len(ne_pristine_matched.loc[
                                                ~ne_pristine_matched.Voltage_aligned.isna()])  # number of points
        # applicable to NE in NE_pristine_matched

        soc_vec_prematching = np.linspace(np.min(df_ne_blended['SOC_aligned']),
                                          np.max(df_ne_blended['SOC_aligned']),
                                          len_non_na_ne_pristine_matched)  # vector across blended NE with same number
        # of applicable points as original (NE_pristine_matched)

        df_ne_blended_matched = ne_pristine_matched.copy()

        df_ne_slice_for_matched = df_ne_blended_interper(soc_vec_prematching)
        df_ne_blended_matched.at[(df_ne_blended_matched.loc[
            (~df_ne_blended_matched['Voltage_aligned'].isna())]).index, 'Voltage_aligned'] = df_ne_slice_for_matched

        return df_ne_blended_matched

    def halfcell_initial_matching_v2(self, x, *params):
        """
        Augments halfcell voltage profiles by scaling and translating them. Typically used in an optimization routine
        to fit the emulated full cell profile to a real cell profile. Alternatively, this function can be used for
        emulation of full cell voltage profiles from its electrode constituents with specified capacity ratio and
        offset of the two electrodes.
        Inputs:
        x: an array of 2 or 3 parameters containing scale_ratio, offset, and optionally NE_2_x. scale_ratio is equal
            to the capacity of the
        cathode divided by the capacity of the anode. offset is defined as the SOC between the cathode at zero capacity
            and the anode at zero
        capacity. NE_2_x is the fraction of the secondary electrode material in the anode.
        df_real: dataframe for the first diagnostic (pristine) of the real full cell. Columns for SOC (evenly spaced)
            and Voltage.
        df_pe: dataframe for the positive electrode. Columns for SOC (evenly spaced) and Voltage.
        self.ne_1_pristine: dataframe for the primary material in the negative electrode. Columns for SOC
            (evenly spaced) and Voltage.
        df_ne_2: dataframe for the secondary material in the negative electrode. Columns for SOC (evenly spaced)
            and Voltage. Supply empty DataFrame if not emulating a blend from two known elelctrodes.
        """

        df_real, df_pe, df_ne_1, df_ne_2_pos, df_ne_2_neg = params
        df_pe = self.pe_pristine
        df_ne_1 = self.ne_1_pristine
        df_ne_2_pos = self.ne_2_pristine_pos
        df_ne_2_neg = self.ne_2_pristine_neg
        scale_ratio = x[0]
        offset = x[1]

        if df_ne_2_pos.empty | df_ne_2_neg.empty:
            # one-material anode
            df_ne = pd.DataFrame()
            df_ne['Voltage_aligned'] = df_ne_1['Voltage_aligned']
            df_ne['SOC_aligned'] = df_ne_1['SOC_aligned']
        else:
            # blended anode
            x_ne_2 = x[2]
            df_ne = blend_electrodes(df_ne_1, df_ne_2_pos, df_ne_2_neg, x_ne_2)  # _robust_v2

        # shifted cathode
        shifted_pe = df_pe.copy()
        shifted_pe['SOC_aligned'] = shifted_pe['SOC_aligned'] * scale_ratio + offset

        # shifted anode
        shifted_ne = df_ne.copy()

        # Interpolate across the max and min SOC of the half-cell dfs
        df_1 = shifted_pe.copy()
        df_2 = shifted_ne.copy()
        min_soc = np.min((np.min(df_1['SOC_aligned']), np.min(df_2['SOC_aligned'])))
        max_soc = np.max((np.max(df_1['SOC_aligned']), np.max(df_2['SOC_aligned'])))
        soc_vec = np.linspace(min_soc, max_soc, 1001)

        df_1_interper = interp1d(df_1['SOC_aligned'],
                                 df_1['Voltage_aligned'],
                                 bounds_error=False, fill_value=np.nan)
        df_1['SOC_aligned'] = soc_vec.copy()
        df_1['Voltage_aligned'] = df_1_interper(soc_vec)

        df_2_interper = interp1d(df_2['SOC_aligned'],
                                 df_2['Voltage_aligned'],
                                 bounds_error=False, fill_value=np.nan)
        df_2['SOC_aligned'] = soc_vec.copy()
        df_2['Voltage_aligned'] = df_2_interper(soc_vec)

        # Calculate the full-cell profile
        df_3 = pd.DataFrame()
        df_3['Voltage_aligned'] = df_1['Voltage_aligned'].subtract(df_2['Voltage_aligned'])
        df_3['SOC_aligned'] = df_2['SOC_aligned']

        # centering
        centering_value = - np.min(df_3['SOC_aligned'].iloc[np.argmin(np.abs(df_3['Voltage_aligned'] - 2.7))])

        emulated_full_cell_centered = df_3.copy()
        emulated_full_cell_centered['SOC_aligned'] = df_3['SOC_aligned'] + centering_value

        pe_out_centered = df_1.copy()
        pe_out_centered['SOC_aligned'] = df_1['SOC_aligned'] + centering_value

        ne_out_centered = df_2.copy()
        ne_out_centered['SOC_aligned'] = df_2['SOC_aligned'] + centering_value

        # Scaling

        emulated_full_cell_centered.loc[(emulated_full_cell_centered['Voltage_aligned'] > self.UPPER_VOLTAGE) | (
                    emulated_full_cell_centered['Voltage_aligned'] < self.LOWER_VOLTAGE)] = np.nan

        scaling_value = np.max(emulated_full_cell_centered['SOC_aligned'].loc[~emulated_full_cell_centered[
            'Voltage_aligned'].isna()])  # value to scale emulated back to 100% SOC

        emulated_full_cell_centered_scaled = emulated_full_cell_centered.copy()
        emulated_full_cell_centered_scaled['SOC_aligned'] = \
            emulated_full_cell_centered['SOC_aligned'] / scaling_value * 100

        pe_out_centered_scaled = pe_out_centered.copy()
        pe_out_centered_scaled['SOC_aligned'] = pe_out_centered['SOC_aligned'] / scaling_value * 100

        ne_out_centered_scaled = ne_out_centered.copy()
        ne_out_centered_scaled['SOC_aligned'] = ne_out_centered['SOC_aligned'] / scaling_value * 100

        # Make new interpolation across SOC for full-cell error calculation
        emulated_full_cell_interper = interp1d(
            emulated_full_cell_centered_scaled.loc[
                ~emulated_full_cell_centered_scaled.Voltage_aligned.isna()].SOC_aligned,
            emulated_full_cell_centered_scaled.loc[
                ~emulated_full_cell_centered_scaled.Voltage_aligned.isna()].Voltage_aligned,
            bounds_error=False, fill_value='extrapolate', kind='quadratic')
        real_full_cell_interper = interp1d(df_real.SOC_aligned,
                                           df_real.Voltage_aligned,
                                           bounds_error=False, fill_value=(self.LOWER_VOLTAGE, self.UPPER_VOLTAGE))

        # Interpolate the emulated full-cell profile
        SOC_vec_full_cell = np.linspace(0, 100.0, 1001)
        emulated_full_cell_interped = pd.DataFrame()
        emulated_full_cell_interped['SOC_aligned'] = SOC_vec_full_cell
        emulated_full_cell_interped['Voltage_aligned'] = emulated_full_cell_interper(SOC_vec_full_cell)

        # Interpolate the true full-cell profile
        df_real_interped = emulated_full_cell_interped.copy()
        df_real_interped['SOC_aligned'] = SOC_vec_full_cell
        df_real_interped['Voltage_aligned'] = real_full_cell_interper(SOC_vec_full_cell)

        return pe_out_centered_scaled, ne_out_centered_scaled, df_real_interped, emulated_full_cell_interped

    def halfcell_initial_matching(self, x, *params):
        """
        Augments halfcell voltage profiles by scaling and translating them. Typically used in an optimization
        routine to fit the emulated full cell profile to a real cell profile.
        Inputs:
        x: an array of 4 or 5 parameters containing scale_pe, offset_pe,scale_pe, scale_ne, offset_ne,
            and optionally ne_2_x
        df_real: dataframe for the first diagnostic (pristine) of the real full cell. Columns for SOC (ev)
        self.pe_pristine: dataframe for the positive electrode. Columns for SOC (evenly spaced) and Voltage.
        df_ne_1: dataframe for the primary material in the negative electrode. Columns for SOC (evenly spaced)
            and Voltage.
        df_ne_2: dataframe for the secondary material in the negative electrode. Columns for SOC (evenly spaced)
            and Voltage. Supply empty DataFrame if not emulating a blend from two known elelctrodes.
        """

        df_real, df_pe, df_ne_1, df_ne_2 = params

        scale_pe = x[0]
        offset_pe = x[1]
        scale_ne = x[2]
        offset_ne = x[3]

        if df_ne_2.empty:
            # one-material anode
            df_ne = pd.DataFrame()
            df_ne['Voltage_aligned'] = df_ne_1['Voltage_aligned']
            df_ne['SOC_aligned'] = df_ne_1['SOC_aligned']
        else:
            # blended anode
            ne_2_x = x[4]  # fraction of NE_2
            df_ne = pd.DataFrame()
            df_ne['Voltage_aligned'] = (ne_2_x * df_ne_2['Voltage_aligned'] + (1 - ne_2_x) * df_ne_1['Voltage_aligned'])
            df_ne['SOC_aligned'] = df_ne_1['SOC_aligned']

        # shifted cathode
        shifted_pe = df_pe.copy()
        shifted_pe['SOC_aligned'] = shifted_pe['SOC_aligned'] * scale_pe + offset_pe

        # shifted anode
        shifted_ne = df_ne.copy()
        shifted_ne['SOC_aligned'] = shifted_ne['SOC_aligned'] * scale_ne + offset_ne

        # Interpolate across the max and min SOC of the half-cell dfs
        df_1 = shifted_pe.copy()
        df_2 = shifted_ne.copy()
        min_soc = np.min((np.min(df_1['SOC_aligned']), np.min(df_2['SOC_aligned'])))
        max_soc = np.max((np.max(df_1['SOC_aligned']), np.max(df_2['SOC_aligned'])))
        soc_vec = np.linspace(min_soc, max_soc, 1001)

        df_1_interper = interp1d(df_1['SOC_aligned'], df_1['Voltage_aligned'], bounds_error=False)
        df_1['SOC_aligned'] = soc_vec.copy()
        df_1['Voltage_aligned'] = df_1_interper(soc_vec)

        df_2_interper = interp1d(df_2['SOC_aligned'], df_2['Voltage_aligned'], bounds_error=False)
        df_2['SOC_aligned'] = soc_vec.copy()
        df_2['Voltage_aligned'] = df_2_interper(soc_vec)

        # Calculate the full-cell profile
        df_3 = pd.DataFrame()
        df_3['Voltage_aligned'] = df_1['Voltage_aligned'].subtract(df_2['Voltage_aligned'])
        df_3.loc[(df_3['Voltage_aligned'] > 4.2) | (df_3['Voltage_aligned'] < 2.7)] = np.nan
        df_3['SOC_aligned'] = df_2['SOC_aligned']

        # centering
        centering_value = (df_real['SOC_aligned'].loc[(np.argmin(np.abs(df_real['Voltage_aligned'] -
                                                                        np.min(df_3['Voltage_aligned'].loc[
                                                                                   ~df_3['Voltage_aligned'].isna()]))))]
                           - np.min(df_3['SOC_aligned'].loc[~df_3['Voltage_aligned'].isna()])
                           )
        emulated_full_cell_centered = df_3.copy()
        emulated_full_cell_centered['SOC_aligned'] = df_3['SOC_aligned'] + centering_value

        PE_out_centered = df_1.copy()
        PE_out_centered['SOC_aligned'] = df_1['SOC_aligned'] + centering_value

        NE_out_centered = df_2.copy()
        NE_out_centered['SOC_aligned'] = df_2['SOC_aligned'] + centering_value

        # Make new interpolation across SOC for full-cell error calculation
        min_soc_full_cell = np.min((np.min(df_3.loc[~df_3.Voltage_aligned.isna()].SOC_aligned),
                                    np.min(df_real.loc[~df_real.Voltage_aligned.isna()].SOC_aligned)))
        max_soc_full_cell = np.max((np.max(df_3.loc[~df_3.Voltage_aligned.isna()].SOC_aligned),
                                    np.max(df_real.loc[~df_real.Voltage_aligned.isna()].SOC_aligned)))
        soc_vec_full_cell = np.linspace(min_soc_full_cell, max_soc_full_cell, 1001)

        emulated_full_cell_interper = interp1d(emulated_full_cell_centered.SOC_aligned,
                                               emulated_full_cell_centered.Voltage_aligned, bounds_error=False)
        real_full_cell_interper = interp1d(df_real.SOC_aligned,
                                           df_real.Voltage_aligned, bounds_error=False)

        # Interpolate the emulated full-cell profile
        emulated_full_cell_interped = pd.DataFrame()
        emulated_full_cell_interped['SOC_aligned'] = soc_vec_full_cell
        emulated_full_cell_interped['Voltage_aligned'] = emulated_full_cell_interper(soc_vec_full_cell)

        # Interpolate the true full-cell profile
        df_real_interped = emulated_full_cell_interped.copy()
        df_real_interped['SOC_aligned'] = soc_vec_full_cell
        df_real_interped['Voltage_aligned'] = real_full_cell_interper(soc_vec_full_cell)

        return df_1, df_2, df_real_interped, emulated_full_cell_interped

    def _get_error_from_halfcell_initial_matching(self, x, *params):
        df_1, df_2, df_real_interped, emulated_full_cell_interped = self.halfcell_initial_matching_v2(x, *params)

        error_dis = distance.euclidean(
            df_real_interped['Voltage_aligned'].loc[
                (~df_real_interped['Voltage_aligned'].isna())
                & (~emulated_full_cell_interped['Voltage_aligned'].isna())],
            emulated_full_cell_interped['Voltage_aligned'].loc[
                (~df_real_interped['Voltage_aligned'].isna())
                & (~emulated_full_cell_interped['Voltage_aligned'].isna())]
                                  )

        error = error_dis + 0.01 * len(emulated_full_cell_interped['Voltage_aligned'].loc[
            emulated_full_cell_interped['Voltage_aligned'].isna()]
                                                 )

        return error

    def _impose_degradation(self,
                            pe_pristine=pd.DataFrame(),
                            ne_1_pristine=pd.DataFrame(),
                            ne_2_pristine_pos=pd.DataFrame(),
                            ne_2_pristine_neg=pd.DataFrame(),
                            lli=0.0, lam_pe=0.0, lam_ne=0.0, x_ne_2=0.0):
        pe_translation = 0
        pe_shrinkage = 0
        ne_translation = 0
        ne_shrinkage = 0

        # Blend negative electrodes
        ne_pristine = blend_electrodes(ne_1_pristine, ne_2_pristine_pos, ne_2_pristine_neg, x_ne_2)

        # Update degradation shifts for LLI
        ne_translation += lli

        # Update degradation shifts for LAM_PE
        upper_voltage_limit = self.UPPER_VOLTAGE
        pe_soc_setpoint = pe_pristine['SOC_aligned'].loc[
            np.argmin(np.abs(pe_pristine['Voltage_aligned']
                             - ne_pristine[
                                 'Voltage_aligned'] - upper_voltage_limit))]  # SOC at which the upper voltage
        # limit is hit
        pe_translation += lam_pe * pe_soc_setpoint / 100  # correction for shrinkage to ensure the locking of
        # the upper voltage SOC
        pe_shrinkage += lam_pe  # shrinkage of PE capacity due to LAM_PE

        #  Update degradation shifts for LAM_NE
        lower_voltage_limit = self.LOWER_VOLTAGE
        ne_soc_setpoint = ne_pristine['SOC_aligned'].loc[
            np.argmin((pe_pristine['Voltage_aligned']
                       - ne_pristine[
                           'Voltage_aligned'] - lower_voltage_limit))]  # SOC at which the lower voltage limit is hit
        ne_translation += lam_ne * ne_soc_setpoint / 100  # correction for shrinkage to ensure the locking of the
        # lower voltage SOC
        ne_shrinkage += lam_ne  # shrinkage of NE capacity due to LAM_NE

        # Update SOC vector for both electrodes according to their imposed degradation
        pe_pristine_shifted_by_deg = pe_pristine.copy()
        pe_pristine_shifted_by_deg['SOC_aligned'] = pe_pristine_shifted_by_deg['SOC_aligned'] * (
                    1 - pe_shrinkage / 100) + pe_translation

        ne_pristine_shifted_by_deg = ne_pristine.copy()
        ne_pristine_shifted_by_deg['SOC_aligned'] = ne_pristine_shifted_by_deg['SOC_aligned'] * (
                    1 - ne_shrinkage / 100) + ne_translation

        # Re-interpolate to align dataframes for differencing
        lower_soc = np.min((np.min(pe_pristine_shifted_by_deg['SOC_aligned']),
                            np.min(ne_pristine_shifted_by_deg['SOC_aligned'])))
        upper_soc = np.max((np.max(pe_pristine_shifted_by_deg['SOC_aligned']),
                            np.max(ne_pristine_shifted_by_deg['SOC_aligned'])))
        soc_vec = np.linspace(lower_soc, upper_soc, 1001)

        pe_pristine_interper = interp1d(pe_pristine_shifted_by_deg['SOC_aligned'],
                                        pe_pristine_shifted_by_deg['Voltage_aligned'], bounds_error=False)
        pe_degraded = pe_pristine_shifted_by_deg.copy()
        pe_degraded['SOC_aligned'] = soc_vec
        pe_degraded['Voltage_aligned'] = pe_pristine_interper(soc_vec)

        ne_pristine_interper = interp1d(ne_pristine_shifted_by_deg['SOC_aligned'],
                                        ne_pristine_shifted_by_deg['Voltage_aligned'], bounds_error=False)
        ne_degraded = ne_pristine_shifted_by_deg.copy()
        ne_degraded['SOC_aligned'] = soc_vec
        ne_degraded['Voltage_aligned'] = ne_pristine_interper(soc_vec)

        return pe_degraded, ne_degraded

    def get_dQdV_over_V_from_degradation_matching(self, x, *params):
        pe_out_centered, ne_out_centered, df_real_interped, emulated_full_cell_interped = \
            self.halfcell_degradation_matching_v3(x, *params)

        # Calculate dQdV from full cell profiles
        dq_dv_real = pd.DataFrame(np.gradient(df_real_interped['SOC_aligned'], df_real_interped['Voltage_aligned']),
                                  columns=['dQdV']).ewm(alpha=x[-2]).mean()
        dq_dv_emulated = pd.DataFrame(
            np.gradient(emulated_full_cell_interped['SOC_aligned'], emulated_full_cell_interped['Voltage_aligned']),
            columns=['dQdV']).ewm(alpha=x[-1]).mean()

        # Include original data
        dq_dv_real['SOC_aligned'] = df_real_interped['SOC_aligned']
        dq_dv_real['Voltage_aligned'] = df_real_interped['Voltage_aligned']

        dq_dv_emulated['SOC_aligned'] = emulated_full_cell_interped['SOC_aligned']
        dq_dv_emulated['Voltage_aligned'] = emulated_full_cell_interped['Voltage_aligned']

        # Interpolate over V
        voltage_vec = np.linspace(2.7, 4.2, 1001)

        v_dq_dv_interper_real = interp1d(dq_dv_real['Voltage_aligned'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                         dq_dv_real['dQdV'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                         bounds_error=False, fill_value=0)
        v_soc_interper_real = interp1d(dq_dv_real['Voltage_aligned'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                       dq_dv_real['SOC_aligned'].loc[~dq_dv_real['Voltage_aligned'].isna()],
                                       bounds_error=False, fill_value=(0, 100))

        v_dq_dv_interper_emulated = interp1d(dq_dv_emulated['Voltage_aligned'].loc[
                                                 ~dq_dv_emulated['Voltage_aligned'].isna()],
                                             dq_dv_emulated['dQdV'].loc[~dq_dv_emulated['Voltage_aligned'].isna()],
                                             bounds_error=False, fill_value=0)
        v_soc_interper_emulated = interp1d(dq_dv_emulated['Voltage_aligned'].loc[
                                               ~dq_dv_emulated['Voltage_aligned'].isna()],
                                           dq_dv_emulated['SOC_aligned'].loc[~dq_dv_emulated['Voltage_aligned'].isna()],
                                           bounds_error=False, fill_value=(0, 100))

        dq_dv_over_v_real = pd.DataFrame(v_dq_dv_interper_real(voltage_vec), columns=['dQdV']).fillna(0)
        dq_dv_over_v_real['SOC_aligned'] = v_soc_interper_real(voltage_vec)
        dq_dv_over_v_real['Voltage_aligned'] = voltage_vec

        dq_dv_over_v_emulated = pd.DataFrame(v_dq_dv_interper_emulated(voltage_vec), columns=['dQdV']).fillna(0)
        dq_dv_over_v_emulated['SOC_aligned'] = v_soc_interper_emulated(voltage_vec)
        dq_dv_over_v_emulated['Voltage_aligned'] = voltage_vec

        return (pe_out_centered,
                ne_out_centered,
                dq_dv_over_v_real,
                dq_dv_over_v_emulated,
                df_real_interped,
                emulated_full_cell_interped)

    def get_error_dQdV_over_V_from_degradation_matching(self, x, *params):
        try:
            (pe_out_centered, ne_out_centered,
             dq_dv_over_v_real, dq_dv_over_v_emulated,
             df_real_interped, emulated_full_cell_interped) = self.get_dQdV_over_V_from_degradation_matching(x, *params)
            error = distance.euclidean(dq_dv_over_v_real['dQdV'], dq_dv_over_v_emulated['dQdV']) + 0.01 * len(
                dq_dv_over_v_emulated['dQdV'].loc[dq_dv_over_v_emulated['dQdV'].isna()])
        except ValueError:
            error = 1000  # set an error for cases where calcuation fails
        return error

    def get_dVdQ_over_SOC_from_degradation_matching(self, x, *params):
        pe_out_centered, ne_out_centered, df_real_interped, emulated_full_cell_interped = \
            self.halfcell_degradation_matching_v3(x, *params)

        # Calculate dVdQ from full cell profiles
        dv_dq_over_soc_real = pd.DataFrame(np.gradient(df_real_interped['Voltage_aligned'],
                                                       df_real_interped['SOC_aligned']),
                                           columns=['dVdQ']).ewm(alpha=x[-2]).mean()
        dv_dq_over_soc_emulated = pd.DataFrame(
            np.gradient(emulated_full_cell_interped['Voltage_aligned'], emulated_full_cell_interped['SOC_aligned']),
            columns=['dVdQ']).ewm(alpha=x[-1]).mean()

        # Include original data
        dv_dq_over_soc_real['SOC_aligned'] = df_real_interped['SOC_aligned']
        dv_dq_over_soc_real['Voltage_aligned'] = df_real_interped['Voltage_aligned']

        dv_dq_over_soc_emulated['SOC_aligned'] = emulated_full_cell_interped['SOC_aligned']
        dv_dq_over_soc_emulated['Voltage_aligned'] = emulated_full_cell_interped['Voltage_aligned']

        # Interpolate over Q
        # ^^ already done in this case as standard data template is over Q

        return (pe_out_centered,
                ne_out_centered,
                dv_dq_over_soc_real,
                dv_dq_over_soc_emulated,
                df_real_interped,
                emulated_full_cell_interped)

    def get_error_dVdQ_over_SOC_from_degradation_matching(self, x, *params):
        try:
            (pe_out_centered, ne_out_centered,
             dv_dq_over_soc_real, dv_dq_over_soc_emulated,
             df_real_interped, emulated_full_cell_interped) = \
                self.get_dVdQ_over_SOC_from_degradation_matching(x, *params)
            error = distance.euclidean(dv_dq_over_soc_real['dVdQ'], dv_dq_over_soc_emulated['dVdQ']) + 0.01 * len(
                dv_dq_over_soc_emulated['dVdQ'].loc[dv_dq_over_soc_emulated['dVdQ'].isna()])
        except ValueError:
            error = 1000  # set an error for cases where calcuation fails
        return error

    def get_V_over_SOC_from_degradation_matching(self, x, *params):
        (PE_out_centered, NE_out_centered, real_aligned, emulated_aligned) = \
            self.halfcell_degradation_matching_v3(x, *params)

        min_soc_full_cell = np.min(real_aligned.loc[~real_aligned.Voltage_aligned.isna()].SOC_aligned)
        max_soc_full_cell = np.max(real_aligned.loc[~real_aligned.Voltage_aligned.isna()].SOC_aligned)

        soc_vec_full_cell = np.linspace(min_soc_full_cell, max_soc_full_cell, 1001)

        emulated_full_cell_interper = interp1d(
            emulated_aligned.SOC_aligned.loc[~real_aligned.Voltage_aligned.isna()],
            emulated_aligned.Voltage_aligned.loc[~real_aligned.Voltage_aligned.isna()],
            bounds_error=False)
        real_full_cell_interper = interp1d(real_aligned.SOC_aligned.loc[~real_aligned.Voltage_aligned.isna()],
                                           real_aligned.Voltage_aligned.loc[~real_aligned.Voltage_aligned.isna()],
                                           bounds_error=False)

        # Interpolate the emulated full-cell profile
        emulated_full_cell_interped = pd.DataFrame()
        emulated_full_cell_interped['SOC_aligned'] = soc_vec_full_cell
        emulated_full_cell_interped['Voltage_aligned'] = emulated_full_cell_interper(soc_vec_full_cell)

        # Interpolate the true full-cell profile
        df_real_interped = emulated_full_cell_interped.copy()
        df_real_interped['SOC_aligned'] = soc_vec_full_cell
        df_real_interped['Voltage_aligned'] = real_full_cell_interper(soc_vec_full_cell)
        return PE_out_centered, NE_out_centered, df_real_interped, emulated_full_cell_interped

    def _get_error_from_degradation_matching(self, x, *params):
        try:
            (pe_out_centered, ne_out_centered, real_aligned, emulated_aligned
             ) = self.get_V_over_SOC_from_degradation_matching(x, *params)
            error = (distance.euclidean(real_aligned.loc[~emulated_aligned.Voltage_aligned.isna()].values.ravel(),
                                        emulated_aligned.loc[~emulated_aligned.Voltage_aligned.isna()].values.ravel()) +
                     0.001 * len(emulated_aligned.loc[emulated_aligned.Voltage_aligned.isna()]))
        except ValueError:
            error = 100
        return error

    def halfcell_degradation_matching_v3(self, x, *params):
        lli = x[0]
        lam_pe = x[1]
        lam_ne = x[2]
        x_ne_2 = x[3]

        pe_pristine, ne_1_pristine, ne_2_pristine_pos, ne_2_pristine_neg, real_full_cell_with_degradation = params

        pe_out, ne_out = self._impose_degradation(pe_pristine, ne_1_pristine,
                                                  ne_2_pristine_pos, ne_2_pristine_neg,
                                                  lli, lam_pe,
                                                  lam_ne, x_ne_2)
        emulated_full_cell_with_degradation = pd.DataFrame()
        emulated_full_cell_with_degradation['SOC_aligned'] = pe_out['SOC_aligned'].copy()
        emulated_full_cell_with_degradation['Voltage_aligned'] = pe_out['Voltage_aligned'] - ne_out['Voltage_aligned']

        emulated_full_cell_with_degradation_centered = pd.DataFrame()

        emulated_full_cell_with_degradation_centered['Voltage_aligned'] = emulated_full_cell_with_degradation[
            'Voltage_aligned']
        centering_value = - np.min(emulated_full_cell_with_degradation['SOC_aligned'].loc[
                                       (~emulated_full_cell_with_degradation['Voltage_aligned'].isna())
                                   ])

        emulated_full_cell_with_degradation_centered['SOC_aligned'] = \
            (emulated_full_cell_with_degradation['SOC_aligned'] + centering_value)

        pe_out_centered = pe_out.copy()
        pe_out_centered['SOC_aligned'] = pe_out['SOC_aligned'] + centering_value

        ne_out_centered = ne_out.copy()
        ne_out_centered['SOC_aligned'] = ne_out['SOC_aligned'] + centering_value

        # Interpolate full profiles across same SOC range
        min_soc = np.min(
            real_full_cell_with_degradation['SOC_aligned'].loc[
                ~real_full_cell_with_degradation['Voltage_aligned'].isna()])
        max_soc = np.max(
            real_full_cell_with_degradation['SOC_aligned'].loc[
                ~real_full_cell_with_degradation['Voltage_aligned'].isna()])
        emulated_interper = interp1d(emulated_full_cell_with_degradation_centered['SOC_aligned'].loc[
                                         ~emulated_full_cell_with_degradation_centered['Voltage_aligned'].isna()],
                                     emulated_full_cell_with_degradation_centered['Voltage_aligned'].loc[
                                         ~emulated_full_cell_with_degradation_centered['Voltage_aligned'].isna()],
                                     bounds_error=False)
        real_interper = interp1d(
            real_full_cell_with_degradation['SOC_aligned'].loc[
                ~real_full_cell_with_degradation['Voltage_aligned'].isna()],
            real_full_cell_with_degradation['Voltage_aligned'].loc[
                ~real_full_cell_with_degradation['Voltage_aligned'].isna()],
            bounds_error=False)

        soc_vec = np.linspace(min_soc, max_soc, 1001)

        emulated_aligned = pd.DataFrame()
        emulated_aligned['SOC_aligned'] = soc_vec
        emulated_aligned['Voltage_aligned'] = emulated_interper(soc_vec)

        real_aligned = pd.DataFrame()
        real_aligned['SOC_aligned'] = soc_vec
        real_aligned['Voltage_aligned'] = real_interper(soc_vec)

        return pe_out_centered, ne_out_centered, real_aligned, emulated_aligned

    def halfcell_degradation_matching_v2(self, x, *params):
        lli = x[0]
        lam_pe = x[1]
        lam_ne = x[2]

        pe_pristine, ne_pristine, real_full_cell_with_degradation = params

        pe_out, ne_out = self._impose_degradation(pe_pristine, ne_pristine, lli, lam_pe, lam_ne)
        emulated_full_cell_with_degradation = pd.DataFrame()
        emulated_full_cell_with_degradation['SOC_aligned'] = pe_out['SOC_aligned'].copy()
        emulated_full_cell_with_degradation['Voltage_aligned'] = pe_out['Voltage_aligned'] - ne_out['Voltage_aligned']

        emulated_full_cell_with_degradation_centered = pd.DataFrame()

        emulated_full_cell_with_degradation_centered['Voltage_aligned'] = emulated_full_cell_with_degradation[
            'Voltage_aligned']
        emulated_full_cell_with_degradation_centered['Voltage_aligned'].loc[
            (emulated_full_cell_with_degradation_centered['Voltage_aligned'] > 4.2) |
            (emulated_full_cell_with_degradation_centered['Voltage_aligned'] < 2.7)] = np.nan

        centering_value = (
                real_full_cell_with_degradation['SOC_aligned'].loc[
                    (np.argmin(np.abs(real_full_cell_with_degradation['Voltage_aligned']
                                      - np.min(emulated_full_cell_with_degradation['Voltage_aligned'].loc[
                                                   ~emulated_full_cell_with_degradation['Voltage_aligned'].isna()]))))
                ]
                - np.min(emulated_full_cell_with_degradation['SOC_aligned'].loc[
                             ~emulated_full_cell_with_degradation['Voltage_aligned'].isna()])
        )

        emulated_full_cell_with_degradation_centered['SOC_aligned'] = \
            (emulated_full_cell_with_degradation['SOC_aligned'] + centering_value)

        pe_out_centered = pe_out.copy()
        pe_out_centered['SOC_aligned'] = pe_out['SOC_aligned'] + centering_value

        ne_out_centered = ne_out.copy()
        ne_out_centered['SOC_aligned'] = ne_out['SOC_aligned'] + centering_value

        # Interpolate full profiles across same SOC range
        min_soc = np.min(
            (np.min(emulated_full_cell_with_degradation_centered['SOC_aligned'].loc[
                        ~emulated_full_cell_with_degradation_centered['Voltage_aligned'].isna()]),
             np.min(real_full_cell_with_degradation['SOC_aligned'].loc[
                        ~real_full_cell_with_degradation['Voltage_aligned'].isna()])
             )
        )
        max_soc = np.max(
            (np.max(emulated_full_cell_with_degradation_centered['SOC_aligned'].loc[
                        ~emulated_full_cell_with_degradation_centered['Voltage_aligned'].isna()]),
             np.max(real_full_cell_with_degradation['SOC_aligned'].loc[
                        ~real_full_cell_with_degradation['Voltage_aligned'].isna()]))
        )
        emulated_interper = interp1d(emulated_full_cell_with_degradation_centered['SOC_aligned'],
                                     emulated_full_cell_with_degradation_centered['Voltage_aligned'],
                                     bounds_error=False)
        real_interper = interp1d(real_full_cell_with_degradation['SOC_aligned'],
                                 real_full_cell_with_degradation['Voltage_aligned'],
                                 bounds_error=False)

        soc_vec = np.linspace(min_soc, max_soc, 1001)

        emulated_aligned = pd.DataFrame()
        emulated_aligned['SOC_aligned'] = soc_vec
        emulated_aligned['Voltage_aligned'] = emulated_interper(soc_vec)

        real_aligned = pd.DataFrame()
        real_aligned['SOC_aligned'] = soc_vec
        real_aligned['Voltage_aligned'] = real_interper(soc_vec)

        return pe_out_centered, ne_out_centered, real_aligned, emulated_aligned

    def get_voltage_curves_for_cell(self, processed_cycler_run, cycle_type='rpt_0.2C'):
        # Filter down to only cycles of cycle_type
        diag_type_cycles = processed_cycler_run.diagnostic_interpolated.loc[
            processed_cycler_run.diagnostic_interpolated['cycle_type'] == cycle_type]

        # Loop across cycles
        cycle_indexes = diag_type_cycles['cycle_index'].unique()
        for i in cycle_indexes:
            diag_type_cycle_i = diag_type_cycles.loc[diag_type_cycles.cycle_index == i]
            x_charge = diag_type_cycle_i.charge_energy.loc[diag_type_cycles['step_type'] == 0]
            y_charge = diag_type_cycle_i.voltage.loc[diag_type_cycles['step_type'] == 0]

        return x_charge, y_charge

    def intracell_wrapper_init(self,
                               cell_struct,
                               initial_matching_bounds=None
                               ):
        """
        Wrapper function that calls all of the functions to get the initial values for the cell
        before fitting all of the
        Args:
             cell_struct (beep.structure): dataframe to determine whether
                charging or discharging
             initial_matching_bounds (tuple): Bounds for fitting parameters
        Returns:
            (bool): True if step is the charge state specified.
        """
        if initial_matching_bounds is None:
            initial_matching_bounds = ((0.8, 1.2), (-20.0, 20.0), (1, 1), (0.1, 0.1), (0.1, 0.1))

        real_cell_initial_charge_profile_aligned, real_cell_initial_charge_profile = \
            self.process_beep_cycle_data_for_initial_halfcell_analysis(cell_struct)
        opt_result_halfcell_initial_matching = differential_evolution(self._get_error_from_halfcell_initial_matching,
                                                                      initial_matching_bounds,
                                                                      args=(real_cell_initial_charge_profile_aligned,
                                                                            self.pe_pristine, self.ne_1_pristine,
                                                                            self.ne_2_pristine_pos,
                                                                            self.ne_2_pristine_neg),
                                                                      strategy='best1bin', maxiter=1000,
                                                                      popsize=15, tol=0.001, mutation=0.5,
                                                                      recombination=0.7, seed=1,
                                                                      callback=None, disp=False, polish=True,
                                                                      init='latinhypercube', atol=0,
                                                                      updating='deferred', workers=-1, constraints=())
        (PE_pristine_matched,
         NE_pristine_matched,
         df_real_interped,
         emulated_full_cell_interped) = self.halfcell_initial_matching_v2(opt_result_halfcell_initial_matching.x,
                                                                          real_cell_initial_charge_profile_aligned,
                                                                          self.pe_pristine,
                                                                          self.ne_1_pristine,
                                                                          self.ne_2_pristine_pos,
                                                                          self.ne_2_pristine_neg)
        return (real_cell_initial_charge_profile_aligned,
                real_cell_initial_charge_profile,
                PE_pristine_matched,
                NE_pristine_matched)

    def intracell_values_wrapper(self,
                                 cycle_index,
                                 cell_struct,
                                 real_cell_initial_charge_profile_aligned,
                                 real_cell_initial_charge_profile,
                                 PE_pristine_matched,
                                 NE_pristine_matched,
                                 degradation_bounds=None
                                 ):

        if degradation_bounds is None:
            degradation_bounds = ((-10, 50),  # LLI
                                  (-10, 50),  # LAM_PE
                                  (-10, 50),  # LAM_NE
                                  (1, 1),  # (-1,1) x_NE_2
                                  (0.1, 0.1),  # (0.01,0.1)
                                  (0.1, 0.1),  # (0.01,0.1)
                                  )

        real_cell_candidate_charge_profile_aligned = self.process_beep_cycle_data_for_candidate_halfcell_analysis(
            cell_struct,
            real_cell_initial_charge_profile_aligned,
            real_cell_initial_charge_profile,
            cycle_index)

        degradation_optimization_result = differential_evolution(self._get_error_from_degradation_matching,
                                                                 degradation_bounds,
                                                                 args=(PE_pristine_matched,
                                                                       NE_pristine_matched,
                                                                       self.ne_2_pristine_pos,
                                                                       self.ne_2_pristine_neg,
                                                                       real_cell_candidate_charge_profile_aligned
                                                                       ),
                                                                 strategy='best1bin', maxiter=100000,
                                                                 popsize=15, tol=0.001, mutation=0.5,
                                                                 recombination=0.7,
                                                                 seed=1,
                                                                 callback=None, disp=False, polish=True,
                                                                 init='latinhypercube',
                                                                 atol=0, updating='deferred', workers=-1,
                                                                 constraints=()
                                                                 )

        (PE_out_centered,
         NE_out_centered,
         dVdQ_over_SOC_real,
         dVdQ_over_SOC_emulated,
         df_real_interped,
         emulated_full_cell_interped) = self.get_dQdV_over_V_from_degradation_matching(
            degradation_optimization_result.x,
            PE_pristine_matched,
            NE_pristine_matched,
            self.ne_2_pristine_pos,
            self.ne_2_pristine_neg,
            real_cell_candidate_charge_profile_aligned)
        #
        (PE_upper_voltage, PE_lower_voltage, PE_upper_SOC, PE_lower_SOC, PE_mass,
         NE_upper_voltage, NE_lower_voltage, NE_upper_SOC, NE_lower_SOC, NE_mass,
         SOC_upper, SOC_lower, Li_mass) = get_halfcell_voltages(PE_out_centered, NE_out_centered)
        #
        LLI = degradation_optimization_result.x[0]
        LAM_PE = degradation_optimization_result.x[1]
        LAM_NE = degradation_optimization_result.x[2]
        x_NE_2 = degradation_optimization_result.x[3]
        alpha_real = degradation_optimization_result.x[4]
        alpha_emulated = degradation_optimization_result.x[5]

        loss_dict = {cycle_index: [LLI, LAM_PE, LAM_NE, x_NE_2, alpha_real, alpha_emulated,
                                   PE_upper_voltage, PE_lower_voltage, PE_upper_SOC, PE_lower_SOC, PE_mass,
                                   NE_upper_voltage, NE_lower_voltage, NE_upper_SOC, NE_lower_SOC, NE_mass,
                                   Li_mass
                                   ]
                     }
        profiles_dict = {cycle_index: real_cell_candidate_charge_profile_aligned}

        return loss_dict, profiles_dict


def blend_electrodes(electrode_1, electrode_2_pos, electrode_2_neg, x_2):
    """
    Inputs:
    electrode_1: Primary material in electrode, typically Gr. DataFrame supplied with SOC evenly spaced and voltage.
    electrode_2: Secondary material in electrode, typically Si. DataFrame supplied with SOC evenly spaced and
        voltage as an additional column.
    x_2: Fraction of electrode_2 material. Supplied as scalar value.
    """
    if electrode_2_pos.empty:
        df_blended = electrode_1
        return df_blended

    if electrode_2_neg.empty:
        electrode_2 = electrode_2_pos
        x_2 = np.abs(x_2)
    elif x_2 > 0:
        electrode_2 = electrode_2_pos
    else:
        electrode_2 = electrode_2_neg
        x_2 = np.abs(x_2)

    electrode_1_interper = interp1d(electrode_1['Voltage_aligned'], electrode_1['SOC_aligned'], bounds_error=False,
                                    fill_value='extrapolate')
    electrode_2_interper = interp1d(electrode_2['Voltage_aligned'], electrode_2['SOC_aligned'], bounds_error=False,
                                    fill_value='extrapolate')

    voltage_vec = np.linspace(np.min((np.min(electrode_1['Voltage_aligned']),
                                      np.min(electrode_2['Voltage_aligned']))),
                              np.max((np.max(electrode_1['Voltage_aligned']),
                                      np.max(electrode_2['Voltage_aligned']))),
                              1001)
    electrode_1_voltage_aligned = pd.DataFrame(electrode_1_interper(voltage_vec), columns=['SOC'])
    electrode_2_voltage_aligned = pd.DataFrame(electrode_2_interper(voltage_vec), columns=['SOC'])
    electrode_1_voltage_aligned['Voltage'] = voltage_vec
    electrode_2_voltage_aligned['Voltage'] = voltage_vec

    df_blend_voltage_aligned = pd.DataFrame(
        (1 - x_2) * electrode_1_voltage_aligned['SOC'] + x_2 * electrode_2_voltage_aligned['SOC'], columns=['SOC'])
    df_blend_voltage_aligned['Voltage'] = electrode_1_voltage_aligned.merge(electrode_2_voltage_aligned,
                                                                            on='Voltage')['Voltage']

    df_blended_interper = interp1d(df_blend_voltage_aligned['SOC'], df_blend_voltage_aligned['Voltage'],
                                   bounds_error=False)
    soc_vec = np.linspace(0, 100, 1001)

    df_blended = pd.DataFrame(df_blended_interper(soc_vec), columns=['Voltage_aligned'])
    df_blended['SOC_aligned'] = soc_vec

    # Modify NE to fully span 100% SOC within its valid voltage window
    df_blended_soc_mod_interper = interp1d(df_blended['SOC_aligned'].loc[~df_blended['Voltage_aligned'].isna()],
                                           df_blended['Voltage_aligned'].loc[~df_blended['Voltage_aligned'].isna()],
                                           bounds_error=False)
    soc_vec = np.linspace(np.min(df_blended['SOC_aligned'].loc[~df_blended['Voltage_aligned'].isna()]),
                          np.max(df_blended['SOC_aligned'].loc[~df_blended['Voltage_aligned'].isna()]),
                          1001)
    df_blended_soc_mod = pd.DataFrame(df_blended_soc_mod_interper(soc_vec), columns=['Voltage_aligned'])
    df_blended_soc_mod['SOC_aligned'] = soc_vec / np.max(soc_vec) * 100
    return df_blended_soc_mod


def get_halfcell_voltages(pe_out_centered, ne_out_centered):
    pe_minus_ne_centered = pd.DataFrame(pe_out_centered['Voltage_aligned'] - ne_out_centered['Voltage_aligned'],
                                        columns=['Voltage_aligned'])
    pe_minus_ne_centered['SOC_aligned'] = pe_out_centered['SOC_aligned']

    soc_upper = pe_minus_ne_centered.iloc[
        np.argmin(np.abs(pe_minus_ne_centered.Voltage_aligned - 4.20))].SOC_aligned
    soc_lower = pe_minus_ne_centered.iloc[
        np.argmin(np.abs(pe_minus_ne_centered.Voltage_aligned - 2.70))].SOC_aligned

    pe_upper_voltage = pe_out_centered.loc[pe_out_centered.SOC_aligned == soc_upper].Voltage_aligned.values[0]
    pe_lower_voltage = pe_out_centered.loc[pe_out_centered.SOC_aligned == soc_lower].Voltage_aligned.values[0]

    pe_upper_soc = ((pe_out_centered.loc[pe_out_centered.Voltage_aligned == pe_upper_voltage]['SOC_aligned'] -
                     np.min(pe_out_centered['SOC_aligned'].loc[~pe_out_centered['Voltage_aligned'].isna()])) / (
                            np.max(pe_out_centered['SOC_aligned'].loc[~pe_out_centered['Voltage_aligned'].isna()]) -
                            np.min(pe_out_centered['SOC_aligned'].loc[~pe_out_centered['Voltage_aligned'].isna()]))
                    ).values[0] * 100
    pe_lower_soc = ((pe_out_centered.loc[pe_out_centered.Voltage_aligned == pe_lower_voltage]['SOC_aligned'] -
                     np.min(pe_out_centered['SOC_aligned'].loc[~pe_out_centered['Voltage_aligned'].isna()])) / (
                            np.max(pe_out_centered['SOC_aligned'].loc[~pe_out_centered['Voltage_aligned'].isna()]) -
                            np.min(pe_out_centered['SOC_aligned'].loc[~pe_out_centered['Voltage_aligned'].isna()]))
                    ).values[0] * 100
    pe_mass = np.max(pe_out_centered['SOC_aligned'].loc[~pe_out_centered['Voltage_aligned'].isna()]) - np.min(
        pe_out_centered['SOC_aligned'].loc[~pe_out_centered['Voltage_aligned'].isna()])

    ne_upper_voltage = ne_out_centered.loc[ne_out_centered.SOC_aligned == soc_upper].Voltage_aligned.values[0]
    ne_lower_voltage = ne_out_centered.loc[ne_out_centered.SOC_aligned == soc_lower].Voltage_aligned.values[0]
    ne_upper_soc = ((ne_out_centered.loc[ne_out_centered.Voltage_aligned == ne_upper_voltage]['SOC_aligned'] -
                     np.min(ne_out_centered['SOC_aligned'].loc[~ne_out_centered['Voltage_aligned'].isna()])) / (
                            np.max(ne_out_centered['SOC_aligned'].loc[~ne_out_centered['Voltage_aligned'].isna()]) -
                            np.min(ne_out_centered['SOC_aligned'].loc[~ne_out_centered['Voltage_aligned'].isna()]))
                    ).values[0] * 100
    ne_lower_soc = ((ne_out_centered.loc[ne_out_centered.Voltage_aligned == ne_lower_voltage]['SOC_aligned'] -
                     np.min(ne_out_centered['SOC_aligned'].loc[~ne_out_centered['Voltage_aligned'].isna()])) / (
                            np.max(ne_out_centered['SOC_aligned'].loc[~ne_out_centered['Voltage_aligned'].isna()]) -
                            np.min(ne_out_centered['SOC_aligned'].loc[~ne_out_centered['Voltage_aligned'].isna()]))
                    ).values[0] * 100
    ne_mass = np.max(ne_out_centered['SOC_aligned'].loc[~ne_out_centered['Voltage_aligned'].isna()]) - np.min(
        ne_out_centered['SOC_aligned'].loc[~ne_out_centered['Voltage_aligned'].isna()])

    li_mass = np.max(pe_minus_ne_centered['SOC_aligned'].loc[~pe_minus_ne_centered.Voltage_aligned.isna()]) - np.min(
        pe_minus_ne_centered['SOC_aligned'].loc[~pe_minus_ne_centered.Voltage_aligned.isna()])

    return (pe_upper_voltage, pe_lower_voltage, pe_upper_soc, pe_lower_soc, pe_mass,
            ne_upper_voltage, ne_lower_voltage, ne_upper_soc, ne_lower_soc, ne_mass,
            soc_upper, soc_lower, li_mass)


def plot_voltage_curves_for_cell(processed_cycler_run,
                                 cycle_type='rpt_0.2C',
                                 x_var='capacity',
                                 step_type=0,
                                 fig_size_inches=(6, 4)):
    # Plot a series of voltage profiles over cycles for this cell
    fig = plt.figure()
    # Filter down to only cycles of cycle_type
    diag_type_cycles = processed_cycler_run.diagnostic_interpolated.loc[
        processed_cycler_run.diagnostic_interpolated['cycle_type'] == cycle_type]

    # Loop across cycles
    cycle_indexes = diag_type_cycles['cycle_index'].unique()
    step_type_list = ['charge', 'discharge']
    for i in cycle_indexes:
        diag_type_cycle_i = diag_type_cycles.loc[diag_type_cycles.cycle_index == i]
        x_charge = diag_type_cycle_i[step_type_list[step_type] + '_' + x_var].loc[
            diag_type_cycles['step_type'] == step_type]
        y_charge = diag_type_cycle_i.voltage.loc[diag_type_cycles['step_type'] == step_type]
        plt.plot(x_charge, y_charge,
                 label='cycle ' + str(i) + ' ' + step_type_list[step_type],
                 c=cm.winter((i - np.min(cycle_indexes)) / (np.max(cycle_indexes) - np.min(cycle_indexes))),
                 figure=fig)

    fig.set_size_inches(fig_size_inches)
    return fig



#  Copyright (c) 2019 Toyota Research Institute

"""
Scripts for profiling memory use for the end to end pipeline.

Usage:
    python memprof.py

Options:
    -h --help        Show this screen
    --version        Show version
"""

import shutil
import os

import boto3

from memory_profiler import profile
from monty.tempfile import ScratchDir
from beep.structure import auto_load
from beep.run_model import DegradationModel
from beep.featurize import DegradationPredictor
from beep import S3_CACHE, tqdm

MEMORY_PROFILE_S3_OBJS = [
    "D3Batt_Data_publication/2017-05-12_5_4C-60per_3_6C_CH23.csv",
    "D3Batt_Data_publication/2017-05-12_5_4C-60per_3_6C_CH23_Metadata.csv",
]


@profile
def memory_profile(s3_objs=None):
    """
    Function for memory profiling pipelines with s3_objs.

    Args:
        s3_objs ([str]): list of s3_objs in the kitware d3batt
            publication s3 bucket.

    """
    s3_objs = s3_objs or MEMORY_PROFILE_S3_OBJS

    # Cache s3 objects
    cache_s3_objs(s3_objs, filter_existing_files=True)

    with ScratchDir("."):
        # Copy all pre-defined s3 objects
        for obj_name in s3_objs:
            shutil.copy(os.path.join(S3_CACHE, obj_name), ".")

        # Snip data path prefix
        data_paths = [os.path.basename(obj) for obj in s3_objs if "Metadata" not in obj]

        # Validation
        # validator = ValidatorBeep()
        # validator.validate_from_paths(data_paths)

        # Data structuring
        datapath_objs = [
            auto_load(data_path) for data_path in data_paths
        ]

        for dp in datapath_objs:
            dp.structure()

        # Featurization
        predictors = [
            DegradationPredictor.init_full_model(dp)
            for dp in datapath_objs
        ]

        # Prediction
        [
            DegradationModel.init_full_model().predict(predictor)
            for predictor in predictors
        ]


def cache_s3_objs(obj_names, bucket_name="kitware", filter_existing_files=True):
    """
    Quick function to download relevant s3 files to cache.

    Args:
        obj_names ([str]): s3 object keys.
        bucket_name (str): bucket name.
        filter_existing_files (bool): whether or not to filter existing files.

    Returns:
        None

    """

    # make cache dir
    if not os.path.isdir(S3_CACHE):
        os.mkdir(S3_CACHE)

    # Initialize s3 resource
    s3 = boto3.resource("s3")
    bucket = s3.Bucket(bucket_name)

    # Put more s3 objects here if desired
    if filter_existing_files:
        obj_names = filter(
            lambda x: not os.path.isfile(os.path.join(S3_CACHE, x)), obj_names
        )
        obj_names = list(obj_names)

    for obj_name in tqdm(obj_names):
        path, filename = os.path.split(obj_name)
        # Make directory if it doesn't exist
        if not os.path.isdir(os.path.join(S3_CACHE, path)):
            os.makedirs(os.path.join(S3_CACHE, path))

        file_object = s3.Object(bucket_name, obj_name)
        file_size = file_object.content_length
        with tqdm(total=file_size, unit_scale=True, desc=filename) as t:
            bucket.download_file(
                obj_name, os.path.join(S3_CACHE, obj_name), Callback=hook(t)
            )


def cache_all_kitware_data():
    """Quick function to cache all of the kitware data."""
    s3 = boto3.client("s3")
    all_objects = s3.list_objects(Bucket="kitware")
    kitware_objects = [
        obj["Key"]
        for obj in all_objects["Contents"]
        if obj["Key"].startswith("D3Batt_Data_publication/")
    ]
    kitware_objects = [obj for obj in kitware_objects if "struct" not in obj]
    kitware_objects.remove("D3Batt_Data_publication/")
    cache_s3_objs(kitware_objects)


def hook(t):
    """tqdm hook for processing s3 downloads."""

    def inner(bytes_amount):
        t.update(bytes_amount)

    return inner


if __name__ == "__main__":
    memory_profile()



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to feature generation"""
import unittest
import os
import numpy as np

from beep.features.core import (
    DeltaQFastCharge,
    CycleSummaryStats,
    TrajectoryFastCharge,
    DiagnosticProperties,
    DiagnosticSummaryStats,
    HPPCResistanceVoltageFeatures

)
from beep.features.base import BEEPFeaturizationError

from beep.structure.maccor import MaccorDatapath
from beep.structure.cli import auto_load_processed, auto_load
from beep.features import featurizer_helpers
from beep.utils import parameters_lookup
from monty.serialization import dumpfn, loadfn
from monty.tempfile import ScratchDir
from beep.utils.s3 import download_s3_object
from beep.tests.constants import TEST_FILE_DIR, BIG_FILE_TESTS, SKIP_MSG


class TestFeaturizer(unittest.TestCase):
    def setUp(self):
        self.structured_cycler_file_path = os.path.join(TEST_FILE_DIR, "2017-06-30_2C-10per_6C_CH10_structure.json")
        self.structured_cycler_file_path_insuf = os.path.join(TEST_FILE_DIR, "structure_insufficient.json")
        self.structured_cycler_file_path_trunc = os.path.join(TEST_FILE_DIR,
                                                              "PreDiag_000240_000227_truncated_structure.json")

    def test_featurization_basic_DeltaQFastCharge(self):
        structured_datapath = auto_load_processed(self.structured_cycler_file_path)
        f = DeltaQFastCharge(structured_datapath)
        self.assertTrue(f.validate()[0])
        f.create_features()

        self.assertEqual(len(f.features), 1)  # just test if works for now
        # Ensure no NaN values
        # print(featurizer.X.to_dict())
        self.assertFalse(np.any(f.features.isnull()))
        self.assertEqual(np.round(f.features.loc[0, 'intercept_discharge_capacity_cycle_number_91:100'], 6),
                         np.round(1.1050065801818196, 6))

        self.assertEqual(f.features.columns.tolist()[4], "charge_time_cycles_1:5")

    def test_feature_serialization(self):
        structured_datapath = auto_load_processed(self.structured_cycler_file_path)

        f = DeltaQFastCharge(structured_datapath)
        f.create_features()

        filename = f.__class__.__name__
        dumpfn(f, filename)

        f_reloaded = loadfn(filename)
        self.assertIsInstance(f_reloaded, DeltaQFastCharge)
        # test nominal capacity is being generated
        self.assertEqual(
            f_reloaded.features.loc[0, "nominal_capacity_by_median"],
            1.0628421000000001
        )

    def test_sequential_multiple_features(self):
        structured_datapath = auto_load_processed(self.structured_cycler_file_path)
        for fclass in [DeltaQFastCharge, TrajectoryFastCharge]:
            f = fclass(structured_datapath)
            self.assertTrue(f.validate()[0])
            f.create_features()

            self.assertEqual(f.metadata["channel_id"], 9)
            self.assertEqual(f.metadata["protocol"], '2017-06-30_tests\\20170629-2C_10per_6C.sdu')
            self.assertEqual(f.metadata["barcode"], 'el150800460605')

            filename = f.__class__.__name__ + ".json"
            dumpfn(f, filename)

        traj = loadfn("TrajectoryFastCharge.json")
        self.assertEqual(traj.features.loc[0, "capacity_0.8"], 161)

    def test_must_fail_featurization(self):
        # insufficient structured file must fail
        structured_insuf = auto_load_processed(self.structured_cycler_file_path_insuf)

        f = DiagnosticSummaryStats(structured_insuf)

        val, msg = f.validate()

        self.assertFalse(val)
        self.assertEqual(msg, "Datapath does not have diagnostic summary")

        unstructured = auto_load(os.path.join(TEST_FILE_DIR, "2017-12-04_4_65C-69per_6C_CH29.csv"))
        with self.assertRaises(BEEPFeaturizationError):
            DiagnosticSummaryStats(unstructured)

    def test_HPPCResistanceVoltageFeatures(self):
        structured_datapath = auto_load_processed(self.structured_cycler_file_path_trunc)
        
        f = HPPCResistanceVoltageFeatures(structured_datapath)
        self.assertTrue(f.validate()[0])
        f.create_features()
        
        self.assertEqual(f.features.shape[1], 76)
        self.assertEqual(f.features.columns[0], "r_c_0s_00")
        self.assertEqual(f.features.columns[-1], "D_8")

        self.assertAlmostEqual(f.features.iloc[0, 0], -0.08845776922490017, 6)
        self.assertAlmostEqual(f.features.iloc[0, 5], -0.1280224700339366, 6)
        self.assertAlmostEqual(f.features.iloc[0, 27], -0.10378359476555565, 6)
        
    def test_DiagnosticSummaryStats(self):
        structured_datapath = auto_load_processed(self.structured_cycler_file_path_trunc)

        f = DiagnosticSummaryStats(structured_datapath)
        self.assertTrue(f.validate()[0])

        f.create_features()

        self.assertEqual(f.features.shape[1], 54)
        self.assertListEqual(
            [f.features.columns[0], f.features.columns[41]],
            ["var_charging_capacity", "square_discharging_dQdV"],
        )
        self.assertListEqual(
            [f.features.columns[42], f.features.columns[53]],
            ["diag_sum_diff_0_1_rpt_0.2Cdischarge_capacity",
             "diag_sum_diff_0_1_rpt_2Ccharge_energy"],
        )
        x = [-3.622991274215596, -1.4948801528128568, -2.441732890889216,
             -0.794422489658189, 0.4889470327970021,
             0.7562360890191123, -0.9122534588595697, -3.771727344982484,
             -1.6613278517299095, -3.9279757071656616,
             0.1418911233780052, 0.7493913209640308, 0.6755655006191633,
             -1.0823827139302122, -2.484906394983077,
             -0.8949449222504844, -1.7523322777749897, -1.4575307327423712,
             0.4467463228405364, 1.3265006178265961,
             0.2422557417274141, -2.6373799375134594, -1.230847957965504,
             -2.046540216421213, 0.2334339752067063,
             0.8239822694093881, 1.2085578295115413, 0.06687710057927358,
             -1.0135736732168983, 0.12101479889802537,
             -2.2735196264247866, 0.37844357940755063, 1.425189114118929,
             1.8786507359201035, 1.6731897281287798,
             -1.1875358619917917, 0.1361208058450041, -1.8275104616090456,
             -0.2665523054105704, 1.1375831683815445,
             1.84972885518774, 1.5023615714170622, -0.00472514151532623,
             -0.003475275535937185,
             -0.008076419207993832, -0.008621551983451683,
             7.413107429038043e-05, 0.0013748657878274915,
             -0.005084993748595586, -0.005675990891556979,
             -0.002536196993382343, -0.0018987653783979423,
             -0.00016598153694586686, -0.00105148083990717]
        computed = f.features.iloc[0].tolist()
        for indx, value in enumerate(x):
            precision = 5
            self.assertEqual(np.around(np.float32(value), precision),
                             np.around(np.float32(computed[indx]), precision))

        self.assertEqual(
            np.around(f.features['var_discharging_capacity'].iloc[0], 6),
            np.around(-3.771727344982484, 6))

        structured_datapath_loc2 = os.path.join(TEST_FILE_DIR,
                                               "PredictionDiagnostics_000136_00002D_truncated_structure.json")
        structured_datapath2 = auto_load_processed(structured_datapath_loc2)

        f2 = DiagnosticSummaryStats(structured_datapath2)
        self.assertTrue(f2.validate()[0])

        f2.create_features()

        x = [-2.4602845133649374, -0.7912059829821004, -1.3246516129064152,
             -0.5577484175221676,
             0.22558675296269257, 1.4107424811304434, 0.44307560772987753,
             -2.968731527885897,
             -1.003386799815887, -1.2861922579124305, 0.010393880890967514,
             0.4995216948726259,
             1.4292366107477192, 0.2643953383205679, -1.3377336978836682,
             -0.21470956778563194,
             -0.7617667690573674, -0.47886877345098366, 0.23547492071796852,
             1.9699615602673914,
             1.566893893282218, -1.8282011110054657, -0.46311299104523346,
             -0.7166620260036703,
             0.06268262404068164, 0.5400910355865228, 2.00139593781454,
             1.4038773986895716,
             0.46799197793006897, 0.5117431282997131, -1.4615182876586914,
             1.2889420237956628,
             2.6205135712205725, 2.176016330718994, 3.1539101600646973,
             -0.9218153953552246,
             0.23360896110534668, -1.1706260442733765, -0.5070897459236073,
             1.1722059184617377,
             2.0029776096343994, 1.7837194204330444, -0.021425815851990795,
             -0.020270314430328763,
             -0.028696091773302315, -0.02782930233422708, -0.017478835661355316,
             -0.019788159842565697,
             -0.021354840746757066, -0.021056601447539146,
             -0.026599426370616085, -0.03017946374275189,
             -0.017983518726387225, -0.01771638489069907]
        computed = f2.features.iloc[0].tolist()
        for indx, value in enumerate(x):
            precision = 5
            self.assertEqual(np.around(np.float32(value), precision),
                             np.around(np.float32(computed[indx]), precision))

    def test_CycleSummaryStats(self):
        structured_datapath_loc = os.path.join(
            TEST_FILE_DIR, "PreDiag_000296_00270E_truncated_structure.json"
        )

        structured_datapath = auto_load_processed(structured_datapath_loc)
        f = CycleSummaryStats(structured_datapath)

        self.assertTrue(f.validate()[0])
        f.create_features()
        self.assertAlmostEqual(f.features['square_discharging_capacity'].iloc[0], 0.764316, 6)
        params_dict = {
            "cycle_comp_num": [11, 100],
            "statistics": ["var", "min", "mean", "skew", "kurtosis", "abs", "square"]
        }
        f2 = CycleSummaryStats(structured_datapath=structured_datapath, hyperparameters=params_dict)
        self.assertTrue(f2.validate()[0])
        f2.create_features()
        self.assertAlmostEqual(f2.features['square_discharging_capacity'].iloc[0], 0.7519596, 6)

    def test_DiagnosticProperties(self):
        structured_datapath_loc = os.path.join(
            TEST_FILE_DIR, "PredictionDiagnostics_000132_00004C_structure.json"
        )
        structured_datapath = auto_load_processed(structured_datapath_loc)

        f = DiagnosticProperties(structured_datapath)
        self.assertTrue(f.validate()[0])
        self.assertListEqual(f.hyperparameters["interpolation_axes"],
                             ['normalized_regular_throughput', 'cycle_index'])
        f.create_features()

        self.assertEqual(f.features.shape, (1, 4))
        # print(list(f.features.iloc[2, :]))

        self.assertAlmostEqual(f.features.iloc[0]["initial_regular_throughput"], 497.587658, 5)

        # Test if rerunning feature works since dictionaries are being updated in place
        f = DiagnosticProperties(structured_datapath)
        f.create_features()
        self.assertEqual(f.features.shape, (1, 4))

class TestFeaturizerHelpers(unittest.TestCase):
    def setUp(self):
        pass

    def test_get_fractional_quantity_remaining_nx(self):
        processed_cycler_run_path_1 = os.path.join(
            TEST_FILE_DIR, "PreDiag_000233_00021F_truncated_structure.json"
        )
        structured_datapath = auto_load_processed(processed_cycler_run_path_1)
        structured_datapath.structured_summary = structured_datapath.structured_summary[
            ~structured_datapath.structured_summary.cycle_index.isin(structured_datapath.diagnostic_summary.cycle_index)]

        sum_diag = featurizer_helpers.get_fractional_quantity_remaining_nx(structured_datapath,
                                                                           metric="discharge_energy",
                                                                           diagnostic_cycle_type="hppc")
        # print(sum_diag["normalized_regular_throughput"])
        self.assertEqual(len(sum_diag.index), 16)
        self.assertEqual(sum_diag.cycle_index.max(), 1507)
        self.assertEqual(np.around(sum_diag["initial_regular_throughput"].iloc[0], 3), np.around(237.001769, 3))
        self.assertEqual(np.around(sum_diag["normalized_regular_throughput"].iloc[15], 3), np.around(45.145, 3))
        self.assertEqual(np.around(sum_diag["normalized_diagnostic_throughput"].iloc[15], 3), np.around(5.098, 3))
        self.assertFalse(sum_diag.isnull().values.any())
        self.assertEqual(sum_diag['diagnostic_start_cycle'].iloc[0], 30)
        self.assertEqual(sum_diag['diagnostic_interval'].iloc[0], 100)
        self.assertEqual(sum_diag['epoch_time'].iloc[0], 1576641695)

        sum_diag = featurizer_helpers.get_fractional_quantity_remaining_nx(structured_datapath,
                                                                           metric="discharge_energy",
                                                                           diagnostic_cycle_type="rpt_1C")
        self.assertEqual(len(sum_diag.index), 16)
        self.assertEqual(sum_diag.cycle_index.max(), 1509)
        self.assertEqual(np.around(sum_diag["initial_regular_throughput"].iloc[0], 3), np.around(237.001769, 3))
        self.assertEqual(np.around(sum_diag["normalized_regular_throughput"].iloc[15], 3), np.around(45.145, 3))
        self.assertEqual(np.around(sum_diag["normalized_diagnostic_throughput"].iloc[15], 3), np.around(5.229, 3))
        self.assertEqual(sum_diag['diagnostic_start_cycle'].iloc[0], 30)
        self.assertEqual(sum_diag['diagnostic_interval'].iloc[0], 100)
        self.assertEqual(sum_diag['epoch_time'].iloc[0], 1576736230)

        processed_cycler_run_path_2 = os.path.join(
            TEST_FILE_DIR, "Talos_001383_NCR18650618001_CH31_truncated_structure.json"
        )
        structured_datapath = auto_load_processed(processed_cycler_run_path_2)

        sum_diag = featurizer_helpers.get_fractional_quantity_remaining_nx(structured_datapath,
                                                                           metric="discharge_energy",
                                                                           diagnostic_cycle_type="hppc")
        self.assertEqual(len(sum_diag.index), 3)
        self.assertEqual(sum_diag.cycle_index.max(), 242)
        self.assertEqual(np.around(sum_diag["initial_regular_throughput"].iloc[0], 3), np.around(331.428, 3))
        self.assertEqual(np.around(sum_diag["normalized_regular_throughput"].iloc[2], 3), np.around(6.817, 3))
        self.assertEqual(np.around(sum_diag["normalized_diagnostic_throughput"].iloc[2], 3), np.around(0.385, 3))
        self.assertEqual(sum_diag['diagnostic_start_cycle'].iloc[0], 30)
        self.assertEqual(sum_diag['diagnostic_interval'].iloc[0], 200)
        self.assertEqual(sum_diag['epoch_time'].iloc[0], 1598156928)

    def test_get_v_diff(self):
        processed_cycler_run_path_1 = os.path.join(
            TEST_FILE_DIR, "Talos_001383_NCR18650618001_CH31_truncated_structure.json"
        )
        processed_cycler_run_path_2 = os.path.join(
            TEST_FILE_DIR, "PreDiag_000304_000153_truncated_structure.json"
        )
        processed_cycler_run_path_3 = os.path.join(
            TEST_FILE_DIR, "Talos_001380_ICR1865026JM001_CH28_truncated_structure.json"
        )
        processed_cycler_run_path_4 = os.path.join(
            TEST_FILE_DIR, "Talos_001375_NCR18650319002_CH15_truncated_structure.json"
        )
        with ScratchDir("."):
            # processed_cycler_run_path_1
            structured_datapath = auto_load_processed(processed_cycler_run_path_1)
            v_vars_df = featurizer_helpers.get_v_diff(structured_datapath, 1, 8)
            # print(v_vars_df)
            self.assertEqual(np.round(v_vars_df.iloc[0]['var_v_diff'], decimals=8),
                             np.round(0.00472705, decimals=8))
            self.assertListEqual(list(v_vars_df.columns),
                                 ["var_v_diff", "min_v_diff", "mean_v_diff", "skew_v_diff", "kurtosis_v_diff",
                                  "sum_v_diff", "sum_square_v_diff"])

            temp_list = v_vars_df.iloc[0, :].to_list()
            temp_list = [np.round(float(x), 8) for x in temp_list]
            self.assertListEqual(temp_list,
                                 [0.00472705, 0.0108896, 0.13865059, 0.59427689, 2.36743208, 176.50219843, 30.4896637])

            # processed_cycler_run_path_2
            structured_datapath = auto_load_processed(processed_cycler_run_path_2)
            v_vars_df = featurizer_helpers.get_v_diff(structured_datapath, 1, 8)
            # print(v_vars_df)
            self.assertEqual(np.round(v_vars_df.iloc[0]['var_v_diff'], decimals=8),
                             np.round(2.664e-05, decimals=8))
            self.assertListEqual(list(v_vars_df.columns),
                                 ["var_v_diff", "min_v_diff", "mean_v_diff", "skew_v_diff", "kurtosis_v_diff",
                                  "sum_v_diff", "sum_square_v_diff"])

            temp_list = v_vars_df.iloc[0, :].to_list()
            temp_list = [np.round(float(x), 8) for x in temp_list]
            self.assertListEqual(temp_list,
                                 [2.664e-05, 0.01481062, 0.01993318, 1.70458503, 4.89453871, 6.83708111, 0.14542267])

            # processed_cycler_run_path_3
            structured_datapath = auto_load_processed(processed_cycler_run_path_3)
            v_vars_df = featurizer_helpers.get_v_diff(structured_datapath, 1, 8)
            # print(v_vars_df)
            self.assertEqual(np.round(v_vars_df.iloc[0]['var_v_diff'], decimals=8),
                             np.round(4.82e-06, decimals=8))
            self.assertListEqual(list(v_vars_df.columns),
                                 ["var_v_diff", "min_v_diff", "mean_v_diff", "skew_v_diff", "kurtosis_v_diff",
                                  "sum_v_diff", "sum_square_v_diff"])

            temp_list = v_vars_df.iloc[0, :].to_list()
            temp_list = [np.round(float(x), 8) for x in temp_list]
            self.assertListEqual(temp_list,
                                 [4.82e-06, 0.01134005, 0.01569094, -0.01052989, 3.25562527, 4.07964428, 0.06526675])

            # processed_cycler_run_path_4
            structured_datapath = auto_load_processed(processed_cycler_run_path_4)
            v_vars_df = featurizer_helpers.get_v_diff(structured_datapath, 1, 8)
            # print(v_vars_df)
            self.assertEqual(np.round(v_vars_df.iloc[0]['var_v_diff'], decimals=8),
                             np.round(9.71e-06, decimals=8))
            self.assertListEqual(list(v_vars_df.columns),
                                 ["var_v_diff", "min_v_diff", "mean_v_diff", "skew_v_diff", "kurtosis_v_diff",
                                  "sum_v_diff", "sum_square_v_diff"])

            temp_list = v_vars_df.iloc[0, :].to_list()
            temp_list = [np.round(float(x), 8) for x in temp_list]
            self.assertListEqual(temp_list,
                                 [9.71e-06, -0.01138431, 0.00490308, -3.09586327, 13.72199015, 2.16744705, 0.01306312])

    def test_get_hppc_ocv(self):
        structured_datapath_loc = os.path.join(
            TEST_FILE_DIR, "PreDiag_000240_000227_truncated_structure.json"
        )
        structured_datapath = auto_load_processed(structured_datapath_loc)
        hppc_ocv_features = featurizer_helpers.get_hppc_ocv(structured_datapath, 1)
        self.assertAlmostEqual(hppc_ocv_features['var_ocv'].iloc[0], 0.000016, 6)
        self.assertAlmostEqual(hppc_ocv_features['min_ocv'].iloc[0], -0.001291, 6)
        self.assertAlmostEqual(hppc_ocv_features['mean_ocv'].iloc[0], 0.002221, 6)
        self.assertAlmostEqual(hppc_ocv_features['skew_ocv'].iloc[0], 1.589392, 6)
        self.assertAlmostEqual(hppc_ocv_features['kurtosis_ocv'].iloc[0], 7.041016, 6)
        self.assertAlmostEqual(hppc_ocv_features['sum_ocv'].iloc[0], 0.025126, 6)
        self.assertAlmostEqual(hppc_ocv_features['sum_square_ocv'].iloc[0], 0.000188, 6)

    def test_get_step_index(self):
        structured_datapath_loc = os.path.join(
            TEST_FILE_DIR, "PreDiag_000240_000227_truncated_structure.json"
        )

        parameters_path = os.path.join(TEST_FILE_DIR, "data-share", "raw", "parameters")
        structured_datapath = auto_load_processed(structured_datapath_loc)
        data = structured_datapath.diagnostic_data
        hppc_cycles = data.loc[data.cycle_type == "hppc"]
        # print(hppc_cycles.step_index.unique())
        _, protocol_name = os.path.split(structured_datapath.metadata.protocol)
        parameter_row, _ = parameters_lookup.get_protocol_parameters(protocol_name, parameters_path=parameters_path)

        for cycle in hppc_cycles.cycle_index.unique():
            hppc_cycle = hppc_cycles[hppc_cycles.cycle_index == cycle]
            for step in hppc_cycle.step_index.unique():
                hppc_cycle_step = hppc_cycle[(hppc_cycle.step_index == step)]
                for step_iter in hppc_cycle_step.step_index_counter.unique():
                    hppc_cycle_step_iter = hppc_cycle_step[(hppc_cycle_step.step_index_counter == step_iter)]
                    duration = hppc_cycle_step_iter.test_time.max() - hppc_cycle_step_iter.test_time.min()
                    median_crate = np.round(hppc_cycle_step.current.median() /
                                            parameter_row["capacity_nominal"].iloc[0], 2)
                    # print(step, median_crate, duration)

        step_ind = featurizer_helpers.get_step_index(structured_datapath,
                                                     cycle_type="hppc",
                                                     diag_pos=0)
        self.assertEqual(len(step_ind.values()), 6)
        print([step_ind["hppc_long_rest"],
               step_ind["hppc_discharge_pulse"],
               step_ind["hppc_short_rest"],
               step_ind["hppc_charge_pulse"],
               step_ind["hppc_discharge_to_next_soc"]])

        self.assertEqual(step_ind, {
            'hppc_charge_to_soc': 9,
            'hppc_long_rest': 11,
            'hppc_discharge_pulse': 12,
            'hppc_short_rest': 13,
            'hppc_charge_pulse': 14,
            'hppc_discharge_to_next_soc': 15
        })
        step_ind = featurizer_helpers.get_step_index(structured_datapath,
                                                     cycle_type="hppc",
                                                     diag_pos=1)
        self.assertEqual(len(step_ind.values()), 6)
        self.assertEqual(step_ind, {
            'hppc_charge_to_soc': 41,
            'hppc_long_rest': 43,
            'hppc_discharge_pulse': 44,
            'hppc_short_rest': 45,
            'hppc_charge_pulse': 46,
            'hppc_discharge_to_next_soc': 47
        })

    def test_get_step_index_2(self):
        structured_datapath_loc = os.path.join(
            TEST_FILE_DIR, "PreDiag_000400_000084_truncated_structure.json"
        )
        parameters_path = os.path.join(TEST_FILE_DIR, "data-share", "raw", "parameters")
        structured_datapath = auto_load_processed(structured_datapath_loc)
        _, protocol_name = os.path.split(structured_datapath.metadata.protocol)
        parameter_row, _ = parameters_lookup.get_protocol_parameters(protocol_name, parameters_path=parameters_path)

        step_ind = featurizer_helpers.get_step_index(structured_datapath,
                                                     cycle_type="hppc",
                                                     diag_pos=0)
        self.assertEqual(len(step_ind.values()), 7)

        self.assertEqual(step_ind, {
            'hppc_charge_to_soc': 9,
            'hppc_long_rest': 11,
            'hppc_discharge_pulse': 12,
            'hppc_short_rest': 13,
            'hppc_charge_pulse': 14,
            'hppc_discharge_to_next_soc': 15,
            'hppc_final_discharge': 17
        })
        step_ind = featurizer_helpers.get_step_index(structured_datapath,
                                                     cycle_type="hppc",
                                                     diag_pos=1)
        self.assertEqual(len(step_ind.values()), 7)
        self.assertEqual(step_ind, {
            'hppc_charge_to_soc': 41,
            'hppc_long_rest': 43,
            'hppc_discharge_pulse': 44,
            'hppc_short_rest': 45,
            'hppc_charge_pulse': 46,
            'hppc_discharge_to_next_soc': 47,
            'hppc_final_discharge': 49
        })
        step_ind = featurizer_helpers.get_step_index(structured_datapath, cycle_type="reset", diag_pos=0)
        self.assertEqual(step_ind, {'reset_charge': 5, 'reset_discharge': 6})
        step_ind = featurizer_helpers.get_step_index(structured_datapath, cycle_type="reset", diag_pos=1)
        self.assertEqual(step_ind, {'reset_charge': 38, 'reset_discharge': 39})

        step_ind = featurizer_helpers.get_step_index(structured_datapath, cycle_type="rpt_0.2C", diag_pos=0)
        self.assertEqual(step_ind, {'rpt_0.2C_charge': 19, 'rpt_0.2C_discharge': 20})
        step_ind = featurizer_helpers.get_step_index(structured_datapath, cycle_type="rpt_0.2C", diag_pos=1)
        self.assertEqual(step_ind, {'rpt_0.2C_charge': 51, 'rpt_0.2C_discharge': 52})

        step_ind = featurizer_helpers.get_step_index(structured_datapath, cycle_type="rpt_1C", diag_pos=0)
        self.assertEqual(step_ind, {'rpt_1C_charge': 22, 'rpt_1C_discharge': 23})
        step_ind = featurizer_helpers.get_step_index(structured_datapath, cycle_type="rpt_1C", diag_pos=1)
        self.assertEqual(step_ind, {'rpt_1C_charge': 54, 'rpt_1C_discharge': 55})

        step_ind = featurizer_helpers.get_step_index(structured_datapath, cycle_type="rpt_2C", diag_pos=0)
        self.assertEqual(step_ind, {'rpt_2C_charge': 25, 'rpt_2C_discharge': 26})
        step_ind = featurizer_helpers.get_step_index(structured_datapath, cycle_type="rpt_2C", diag_pos=1)
        self.assertEqual(step_ind, {'rpt_2C_charge': 57, 'rpt_2C_discharge': 58})

    def test_get_step_index_3(self):
        structured_datapath_loc = os.path.join(
            TEST_FILE_DIR, "PredictionDiagnostics_000136_00002D_truncated_structure.json"
        )
        structured_datapath = auto_load_processed(structured_datapath_loc)
        step_ind = featurizer_helpers.get_step_index(structured_datapath, cycle_type="hppc", diag_pos=0)
        self.assertEqual(len(step_ind.values()), 6)

    def test_get_diffusion_coeff(self):
        with ScratchDir("."):
            structured_datapath_loc = os.path.join(
                TEST_FILE_DIR, "PreDiag_000240_000227_truncated_structure.json"
            )
            structured_datapath = auto_load_processed(structured_datapath_loc)
            diffusion_df = featurizer_helpers.get_diffusion_coeff(structured_datapath, 1)
            print(np.round(diffusion_df.iloc[0].to_list(), 3))
            self.assertEqual(np.round(diffusion_df.iloc[0].to_list(), 3)[0], -0.016)
            self.assertEqual(np.round(diffusion_df.iloc[0].to_list(), 3)[5], -0.011)


class TestRawToFeatures(unittest.TestCase):
    def setUp(self):
        self.maccor_file_w_parameters_s3 = {
            "bucket": "beep-sync-test-stage",
            "key": "big_file_tests/PreDiag_000287_000128.092"
        }
        self.maccor_file_w_parameters = os.path.join(
            TEST_FILE_DIR, "PreDiag_000287_000128.092"
        )

    @unittest.skipUnless(BIG_FILE_TESTS, SKIP_MSG)
    def test_raw_to_features(self):

        download_s3_object(bucket=self.maccor_file_w_parameters_s3["bucket"],
                           key=self.maccor_file_w_parameters_s3["key"],
                           destination_path=self.maccor_file_w_parameters)

        dp = MaccorDatapath.from_file(self.maccor_file_w_parameters)
        dp.autostructure()
        processed_run_path = os.path.join(
            TEST_FILE_DIR, "processed_diagnostic.json"
        )

        # Dump to the structured file and check the file size
        dumpfn(dp, processed_run_path)

        dp = loadfn(processed_run_path)

        for fclass in (
                DeltaQFastCharge,
                CycleSummaryStats,
                TrajectoryFastCharge,
                DiagnosticSummaryStats,
                HPPCResistanceVoltageFeatures,
        ):
            f = fclass(dp)
            self.assertTrue(f.validate()[0])
            f.create_features()

            if fclass.__class__.__name__ == "HPPCResistanceVoltageFeatures":
                self.assertAlmostEqual(f.features['r_c_0s_00'].iloc[0], -0.159771397, 5)
                self.assertAlmostEqual(f.features['r_c_0s_10'].iloc[0], -0.143679, 5)
                self.assertAlmostEqual(f.features['r_c_0s_20'].iloc[0], -0.146345, 5)
                self.assertAlmostEqual(f.features['D_6'].iloc[0], -0.167919, 5)
                self.assertAlmostEqual(f.features['D_7'].iloc[0], 0.094136, 5)
                self.assertAlmostEqual(f.features['D_8'].iloc[0], 0.172496, 5)



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to Splicing files"""

import os
import unittest

import numpy as np

from beep.utils import MaccorSplice
from beep.utils.s3 import download_s3_object
from beep import MODULE_DIR
from beep.utils import parameters_lookup
from beep.tests.constants import BIG_FILE_TESTS, TEST_FILE_DIR, SKIP_MSG
from beep import PROTOCOL_PARAMETERS_DIR


class SpliceTest(unittest.TestCase):
    def setUp(self):
        self.arbin_file = os.path.join(TEST_FILE_DIR, "FastCharge_000000_CH29.csv")
        self.filename_part_1 = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000038.078")
        self.filename_part_2 = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000038con.078")
        self.output = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000038joined.078")
        self.test = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000038test.078")

    def test_maccor_read_write(self):
        splicer = MaccorSplice(self.filename_part_1, self.filename_part_2, self.output)

        meta_1, data_1 = splicer.read_maccor_file(self.filename_part_1)
        splicer.write_maccor_file(meta_1, data_1, self.test)
        meta_test, data_test = splicer.read_maccor_file(self.test)

        assert meta_1 == meta_test
        assert np.allclose(data_1["Volts"].to_numpy(), data_test["Volts"].to_numpy())
        assert np.allclose(data_1["Amps"].to_numpy(), data_test["Amps"].to_numpy())
        assert np.allclose(
            data_1["Test (Sec)"].to_numpy(), data_test["Test (Sec)"].to_numpy()
        )

    def test_column_increment(self):
        splicer = MaccorSplice(self.filename_part_1, self.filename_part_2, self.output)
        meta_1, data_1 = splicer.read_maccor_file(self.filename_part_1)
        meta_2, data_2 = splicer.read_maccor_file(self.filename_part_2)
        data_1, data_2 = splicer.column_increment(data_1, data_2)

        assert data_1["Rec#"].max() < data_2["Rec#"].min()


@unittest.skipUnless(BIG_FILE_TESTS, SKIP_MSG)
class S3Test(unittest.TestCase):

    bucket = "beep-sync-test-stage"
    key = "test_util/test_file.txt"
    destination_path = "test_util_s3_file.txt"

    def test_download_s3_object(self):

        download_s3_object(bucket=self.bucket,
                           key=self.key,
                           destination_path=self.destination_path)

        os._exists(self.destination_path)

    def tearDown(self) -> None:
        os.remove(self.destination_path)


class TestStructuringUtils(unittest.TestCase):
    """
    Tests related to utils only used in structuring.
    """

    # based on RCRT.test_get_protocol_parameters
    def test_get_protocol_parameters(self):
        filepath = os.path.join(
            TEST_FILE_DIR, "PredictionDiagnostics_000109_tztest.010"
        )
        parameters, _ = parameters_lookup.get_protocol_parameters(filepath, parameters_path=PROTOCOL_PARAMETERS_DIR)

        self.assertEqual(parameters["diagnostic_type"].iloc[0], "HPPC+RPT")
        self.assertEqual(parameters["diagnostic_parameter_set"].iloc[0], "Tesla21700")
        self.assertEqual(parameters["seq_num"].iloc[0], 109)
        self.assertEqual(len(parameters.index), 1)

        parameters_missing, project_missing = parameters_lookup.get_protocol_parameters(
            "Fake", parameters_path=PROTOCOL_PARAMETERS_DIR
        )
        self.assertEqual(parameters_missing, None)
        self.assertEqual(project_missing, None)

        filepath = os.path.join(TEST_FILE_DIR, "PreDiag_000292_tztest.010")
        parameters, _ = parameters_lookup.get_protocol_parameters(filepath, parameters_path=PROTOCOL_PARAMETERS_DIR)
        self.assertEqual(parameters["diagnostic_type"].iloc[0], "HPPC+RPT")
        self.assertEqual(parameters["seq_num"].iloc[0], 292)

    # based on RCRT.test_get_project_name
    def test_get_project_name(self):
        project_name_parts = parameters_lookup.get_project_sequence(
            os.path.join(TEST_FILE_DIR, "PredictionDiagnostics_000109_tztest.010")
        )
        project_name = project_name_parts[0]
        self.assertEqual(project_name, "PredictionDiagnostics")

    # based on RCRT.test_get_diagnostic_parameters
    def test_get_diagnostic_parameters(self):
        diagnostic_available = {
            "parameter_set": "Tesla21700",
            "cycle_type": ["reset", "hppc", "rpt_0.2C", "rpt_1C", "rpt_2C"],
            "length": 5,
            "diagnostic_starts_at": [1, 36, 141],
        }
        diagnostic_parameter_path = os.path.join(MODULE_DIR, "procedure_templates")
        project_name = "PreDiag"
        v_range = parameters_lookup.get_diagnostic_parameters(
            diagnostic_available, diagnostic_parameter_path, project_name
        )
        self.assertEqual(v_range, [2.7, 4.2])



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for maccor protcol files to biologic modulo bat protcol files"""

import os
from typing import OrderedDict
import unittest
import xmltodict
import copy
import pandas as pd
from monty.tempfile import ScratchDir
from pydash import get
from beep.protocol import (
    PROTOCOL_SCHEMA_DIR,
    BIOLOGIC_TEMPLATE_DIR,
    PROCEDURE_TEMPLATE_DIR,
)
from beep.protocol.maccor import Procedure
from beep.protocol.maccor_to_biologic_mb import (
    MaccorToBiologicMb,
    CycleAdvancementRules,
    CycleAdvancementRulesSerializer
)
from beep.tests.constants import TEST_FILE_DIR


class ConversionTest(unittest.TestCase):
    maxDiff = None

    def maccor_values_to_biologic_value_and_unit_test(self, func, tests):
        for value_str, expected_value_str, expected_unit in tests:
            actual_value, actual_unit = func(value_str)            
            self.assertEqual(actual_value, expected_value_str)
            self.assertEqual(actual_unit, expected_unit)

    def test_convert_volts(self):
        converter = MaccorToBiologicMb()
        tests = [
            ("0.1429", "142.900", "mV"),
            ("0.1429e3", "142.900", "V"),
            ("159.3624", "159362.400", "mV"),
            ("152.9", "152.900",  "V")
        ]
        self.maccor_values_to_biologic_value_and_unit_test(
            converter._convert_volts,
            tests,
        )
    
    def test_convert_amps(self):
        converter = MaccorToBiologicMb()
        tests = [
            ("0.1429", "142.900", "mA"),
            ("1.23", "1.230", "A"),
            ("152.9", "152.900",  "A"),
            ("1.2e-4", "120.000", "\N{Micro Sign}A")
        ]
        self.maccor_values_to_biologic_value_and_unit_test(
            converter._convert_amps,
            tests,
        )

    def test_convert_watts(self):
        converter = MaccorToBiologicMb()
        tests = [
            ("0.1429", "142.900", "mW"),
            ("1.23", "1.230", "W"),
            ("152.9", "152.900",  "W"),
            ("1.2e-5", "12.000", "\N{Micro Sign}W")
        ]
        self.maccor_values_to_biologic_value_and_unit_test(
            converter._convert_watts,
            tests,
        )

    def test_convert_ohms(self):
        converter = MaccorToBiologicMb()
        tests = [
            ("0.1429", "142.900", "mOhms"),
            ("1.459e4", "14.590", "kOhms"),
            ("152.9", "152.900",  "Ohms"),
            ("1.2e-4", "120.000", "\N{Micro Sign}Ohms")
        ]
        self.maccor_values_to_biologic_value_and_unit_test(
            converter._convert_ohms,
            tests,
        )
    
    def test_convert_time(self):
        converter = MaccorToBiologicMb()
        tests = [
            ("::.01", "10.000", "ms"),
            ("03::", "3.000", "h"),
            ("03:30:", "210.000",  "mn"),
            ("00:00:50", "50.000", "s")
        ]
        self.maccor_values_to_biologic_value_and_unit_test(
            converter._convert_time,
            tests,
        )

    def single_step_to_single_seq_test(self, test_step_xml, diff_dict):
        """
        test utility for testing proc_step_to_seq
         """
        proc = xmltodict.parse(test_step_xml)
        test_step = proc["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        converter = MaccorToBiologicMb()

        expected = converter._blank_seq.copy()
        expected["Ns"] = 0
        expected["lim1_seq"] = 1
        expected["lim2_seq"] = 1
        expected["lim3_seq"] = 1
        expected.update(diff_dict)

        step_num = 1
        seq_nums_by_step_num = {
            step_num: [0],
            step_num + 1: [1],
        }

        result = converter._convert_step_parts(
            step_parts=[test_step],
            step_num=step_num,
            seq_nums_by_step_num=seq_nums_by_step_num,
            goto_lowerbound=0,
            goto_upperbound=3,
            end_step_num=4,
        )[0]

        for key, value in expected.items():
            self.assertEqual(
                value,
                result[key],
                msg="Expected {0}: {1} got {0}: {2}".format(key, value, result[key]),
            )

    def test_partition_steps_into_techniques(self):
        converter = MaccorToBiologicMb()
        ast = converter.load_maccor_ast(
            os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV5.000")
        )
        steps = get(ast, "MaccorTestProcedure.ProcSteps.TestStep")
        self.assertEqual(True, len(steps) > 71)
        
        # existence of looped tech 2
        nested_loop_open_idx = 36
        nested_loop_open_type = get(steps[nested_loop_open_idx], 'StepType')
        self.assertEqual(nested_loop_open_type, "Do 1")
        nested_loop_close_idx = 68
        nested_loop_close_type = get(steps[nested_loop_close_idx], 'StepType')
        self.assertEqual(nested_loop_close_type, "Loop 1")


        technique_partitions = converter._partition_steps_into_techniques(steps)
        self.assertEqual(3, len(technique_partitions))
        partition1, partition2, partition3 =  technique_partitions
        
        self.assertEqual(partition1.technique_num, 1)
        self.assertEqual(partition2.technique_num, 2)
        self.assertEqual(partition3.technique_num, 4)

        self.assertEqual(partition1.tech_does_loop, False)
        self.assertEqual(partition2.tech_does_loop, True)
        self.assertEqual(partition3.tech_does_loop, False)

        self.assertEqual(partition2.num_loops, 1000)

        self.assertEqual(partition1.step_num_offset, 0)
        self.assertEqual(partition2.step_num_offset, nested_loop_open_idx + 1)
        self.assertEqual(partition3.step_num_offset, nested_loop_close_idx + 1)

        self.assertEqual(len(partition1.steps), 36)
        # trim opening/closing loops
        self.assertEqual(len(partition2.steps), nested_loop_close_idx - nested_loop_open_idx - 1)
        self.assertEqual(len(partition3.steps), 27)

    def test_apply_step_mappings_global_noop(self):
        xml = (step_with_bounds_template).format(
            voltage_v_lowerbound = 2.2,
            voltage_v_upperbound = 4.2,
            current_a_lowerbound = 0.1,
            current_a_upperbound = 1.0,
        )
        step = get(
            xmltodict.parse(xml, process_namespaces=False, strip_whitespace=True),
            'TestStep',
        )
        converter = MaccorToBiologicMb()
        
        # no limits no mappings
        unmapped_step = converter._apply_step_mappings([step])[0]
        self.assertEqual(step, unmapped_step)

        # limits outside of bounds, don't
        converter.max_voltage_v = 10.0
        converter.min_voltage_v = -10.0
        converter.max_current_a = 10.0
        converter.min_current_a = -10.0
        unmapped_step = converter._apply_step_mappings([step])[0]
        self.assertEqual(step, unmapped_step)
    
    def test_apply_step_mappings_global_voltage(self):
        xml = (step_with_bounds_template).format(
            voltage_v_lowerbound = 2.2,
            voltage_v_upperbound = 4.2,
            current_a_lowerbound = 0.1,
            current_a_upperbound = 1.0,
        )
        step = get(
            xmltodict.parse(xml, process_namespaces=False, strip_whitespace=True),
            'TestStep',
        )
        converter = MaccorToBiologicMb()

        converter.max_voltage_v = 3.9
        converter.min_voltage_v = 3.1
        step_without_voltage_end_entries = converter._apply_step_mappings([step])[0]

        end_entries = get(step_without_voltage_end_entries, "Ends.EndEntry")
        self.assertEqual(2, len(end_entries))
        self.assertEqual("Current", get(end_entries[0], "EndType"))
        self.assertEqual("Current", get(end_entries[1], "EndType"))

        # check there was not mutation
        original_end_entries = get(step, 'Ends.EndEntry')
        self.assertEqual(4, len(original_end_entries))
    
    def test_apply_step_mappings_all_global_limits(self):
        xml = (step_with_bounds_template).format(
            voltage_v_lowerbound = 2.2,
            voltage_v_upperbound = 4.2,
            current_a_lowerbound = 0.1,
            current_a_upperbound = 1.0,
        )
        step = get(
            xmltodict.parse(xml, process_namespaces=False, strip_whitespace=True),
            'TestStep',
        )
        converter = MaccorToBiologicMb()


        converter.max_voltage_v = 3.9
        converter.min_voltage_v = 3.1
        converter.max_current_a = 0.7
        converter.min_current_a = 0.3

        step_with_no_end_entries = converter._apply_step_mappings([step])[0]
        self.assertEqual(None, get(step_with_no_end_entries, "Ends.EndEntry"))

        # check there was not mutation
        original_end_entries = get(step, 'Ends.EndEntry')
        self.assertEqual(4, len(original_end_entries))

    def test_rest_step_conversion(self):
        xml = (
            '<?xml version="1.0" encoding="UTF-8"?>'
            "<MaccorTestProcedure>"
            "  <ProcSteps>"
            "    <TestStep>"
            "      <StepType>  Rest  </StepType>"
            "      <StepMode>        </StepMode>"
            "      <StepValue></StepValue>"
            "      <Limits/>"
            "      <Ends>"
            "        <EndEntry>"
            "          <EndType>Voltage </EndType>"
            "          <SpecialType> </SpecialType>"
            "          <Oper>&gt;= </Oper>"
            "          <Step>002</Step>"
            "          <Value>4.4</Value>"
            "        </EndEntry>"
            "        <EndEntry>"
            "          <EndType>Voltage </EndType>"
            "          <SpecialType> </SpecialType>"
            "          <Oper>&lt;= </Oper>"
            "          <Step>002</Step>"
            "          <Value>2.5</Value>"
            "        </EndEntry>"
            "      </Ends>"
            "      <Reports>"
            "        <ReportEntry>"
            "          <ReportType>Voltage</ReportType>"
            "          <Value>2.2</Value>"
            "        </ReportEntry>"
            "      </Reports>"
            "      <Range>A</Range>"
            "      <Option1>N</Option1>"
            "      <Option2>N</Option2>"
            "      <Option3>N</Option3>"
            "      <StepNote></StepNote>"
            "    </TestStep>"
            "  </ProcSteps>"
            "</MaccorTestProcedure>"
        )
        diff_dict = {
            "ctrl_type": "Rest",
            "Apply I/C": "I",
            "N": "1.00",
            "charge/discharge": "Charge",
            "lim_nb": 2,
            "lim1_type": "Ecell",
            "lim1_comp": ">",
            "lim1_value": "4.400",
            "lim1_value_unit": "V",
            "lim2_type": "Ecell",
            "lim2_comp": "<",
            "lim2_value": "2.500",
            "lim2_value_unit": "V",
            "rec_nb": 1,
            "rec1_type": "Ecell",
            "rec1_value": "2.200",
            "rec1_value_unit": "V",
        }

        self.single_step_to_single_seq_test(xml, diff_dict)
        pass

    def test_discharge_current_step_conversion(self):
        xml = (
            '<?xml version="1.0" encoding="UTF-8"?>'
            "<MaccorTestProcedure>"
            "  <ProcSteps>"
            "    <TestStep>"
            #      mispelling taken directly from sample file
            "      <StepType>Dischrge</StepType>"
            "      <StepMode>Current </StepMode>"
            "      <StepValue>1.0</StepValue>"
            "      <Limits/>"
            "      <Ends>"
            "        <EndEntry>"
            "          <EndType>StepTime</EndType>"
            "          <SpecialType> </SpecialType>"
            "          <Oper> = </Oper>"
            "          <Step>002</Step>"
            "          <Value>00:00:30</Value>"
            "        </EndEntry>"
            "        <EndEntry>"
            "          <EndType>Voltage </EndType>"
            "          <SpecialType> </SpecialType>"
            "          <Oper>&lt;= </Oper>"
            "          <Step>002</Step>"
            "          <Value>2.7</Value>"
            "        </EndEntry>"
            "        <EndEntry>"
            "          <EndType>Voltage </EndType>"
            "          <SpecialType> </SpecialType>"
            "          <Oper>&gt;= </Oper>"
            "          <Step>002</Step>"
            "          <Value>4.4</Value>"
            "        </EndEntry>"
            "      </Ends>"
            "      <Reports>"
            "        <ReportEntry>"
            "          <ReportType>Voltage </ReportType>"
            "          <Value>0.001</Value>"
            "        </ReportEntry>"
            "        <ReportEntry>"
            "          <ReportType>StepTime</ReportType>"
            #          10ms
            "          <Value>::.01</Value>"
            "        </ReportEntry>"
            "      </Reports>"
            "      <Range>A</Range>"
            "      <Option1>N</Option1>"
            "      <Option2>N</Option2>"
            "      <Option3>N</Option3>"
            "      <StepNote></StepNote>"
            "    </TestStep>"
            "  </ProcSteps>"
            "</MaccorTestProcedure>"
        )
        diff_dict = {
            "ctrl_type": "CC",
            "Apply I/C": "I",
            "ctrl1_val": "1.000",
            "ctrl1_val_unit": "A",
            "ctrl1_val_vs": "<None>",
            "N": "15.00",
            "charge/discharge": "Discharge",
            "lim_nb": 3,
            "lim1_type": "Time",
            "lim1_comp": ">",
            "lim1_value": "30.000",
            "lim1_value_unit": "s",
            "lim2_type": "Ecell",
            "lim2_comp": "<",
            "lim2_value": "2.700",
            "lim2_value_unit": "V",
            "lim3_type": "Ecell",
            "lim3_comp": ">",
            "lim3_value": "4.400",
            "lim3_value_unit": "V",
            "rec_nb": 2,
            "rec1_type": "Ecell",
            "rec1_value": "1.000",
            "rec1_value_unit": "mV",
            "rec2_type": "Time",
            "rec2_value": "10.000",
            "rec2_value_unit": "ms",
        }

        self.single_step_to_single_seq_test(xml, diff_dict)
        pass

    def test_conversion_with_updated(self):
        converter = MaccorToBiologicMb()

        with ScratchDir(".") as scratch_dir:
            # Generate a protocol that can be used with the existing cells for testing purposes
            reg_params = {
                'project_name': {0: 'FormDegrade'},
                'seq_num': {0: 0},
                'template': {0: 'diagnosticV5.000'},
                'charge_constant_current_1': {0: 2.0},
                'charge_percent_limit_1': {0: 30},
                'charge_constant_current_2': {0: 2.0},
                'charge_cutoff_voltage': {0: 4.4},
                'charge_constant_voltage_time': {0: 60},
                'charge_rest_time': {0: 5},
                'discharge_constant_current': {0: 1.0},
                'discharge_cutoff_voltage': {0: 3.0},
                'discharge_rest_time': {0: 15},
                'cell_temperature_nominal': {0: 25},
                'cell_type': {0: 'LiFun240'},
                'capacity_nominal': {0: 0.240},
                'diagnostic_type': {0: 'HPPC+RPT'},
                'diagnostic_parameter_set': {0: 'LiFunForm'},
                'diagnostic_start_cycle': {0: 30},
                'diagnostic_interval': {0: 100}
                          }

            protocol_params_df = pd.DataFrame.from_dict(reg_params)
            index = 0
            protocol_params = protocol_params_df.iloc[index]
            diag_params_df = pd.read_csv(
                os.path.join(PROCEDURE_TEMPLATE_DIR, "PreDiag_parameters - DP.csv")
            )
            diagnostic_params = diag_params_df[
                diag_params_df["diagnostic_parameter_set"]
                == protocol_params["diagnostic_parameter_set"]
                ].squeeze()

            procedure = Procedure.generate_procedure_regcyclev3(index, protocol_params)
            procedure.generate_procedure_diagcyclev3(
                protocol_params["capacity_nominal"], diagnostic_params
            )
            procedure.set_skip_to_end_diagnostic(4.5, 2.0, step_key="070", new_step_key="095")
            procedure.to_file(os.path.join(scratch_dir, "BioTest_000001.000"))

            # Setup the converter and run it
            def set_i_range(tech_num, seq, idx):
                seq_copy = copy.deepcopy(seq)
                seq_copy["I Range"] = "1 A"
                return seq_copy
            converter.seq_mappers.append(set_i_range)
            converter.min_voltage_v = 2.0
            converter.max_voltage_v = 4.5

            converter.convert(os.path.join(scratch_dir, "BioTest_000001.000"),
                              TEST_FILE_DIR, "BioTest_000001")
            f = open(os.path.join(TEST_FILE_DIR, "BioTest_000001.mps"), encoding="ISO-8859-1")
            file = f.readlines()
            control_list = [
                'ctrl_type', 'Rest', 'CC', 'Rest', 'CC', 'CV', 'CC', 'Loop', 'CC', 'CV', 'Rest', 'CC',
                'Rest', 'CC', 'CC', 'Loop', 'CV', 'CC', 'CC', 'CV', 'CC', 'CC', 'CV', 'CC', 'CC', 'CV',
                'CC', 'CC', 'CC', 'CV', 'Rest', 'CC', 'Rest', 'Loop'
            ]
            self.assertListEqual(control_list, file[35].split())
            value_list = [
                'ctrl1_val', '240.000', '34.300', '4.400', '34.300', '100.000', '80.000', '4.400', '240.000',
                '180.000', '80.000', '100.000', '3.000', '80.000', '48.000', '4.400', '48.000', '48.000', '4.400',
                '240.000', '48.000', '4.400', '480.000', '480.000', '480.000', '4.400', '240.000', '100.000'
            ]
            self.assertListEqual(value_list, file[37].split())

            voltage_min = '\tEcell min = 2.00 V\n'
            self.assertEqual(voltage_min, file[9])
            voltage_max = '\tEcell max = 4.50 V\n'
            self.assertEqual(voltage_max, file[10])

    def test_cycle_transition_serialization(self):
        cycle_transition_rules = CycleAdvancementRules(
            tech_num=2,
            tech_does_loop=True,
            adv_cycle_on_start=1,
            adv_cycle_on_tech_loop=1,
            adv_cycle_seq_transitions={(2, 5): 1, (14, 17): 1},
            debug_adv_cycle_on_step_transitions={(72, 71): 1, (72, 75): 1},
        )

        serializer = CycleAdvancementRulesSerializer()
        json_str = serializer.json(cycle_transition_rules)
        parsed_cycle_transition_rules = serializer.parse_json(json_str)

        self.assertEqual(
            cycle_transition_rules.__repr__(),
            parsed_cycle_transition_rules.__repr__(),
        )

    def test_convert_protocol_with_goto_end_step(self):
        path = os.path.join(TEST_FILE_DIR, "goto_end_example.000")
        converter = MaccorToBiologicMb()
        ast = converter.load_maccor_ast(path)
        test_steps = get(ast, "MaccorTestProcedure.ProcSteps.TestStep")
        assert isinstance(test_steps, list)

        first_step = test_steps[0]
        first_step_goto = int(get(first_step, "Ends.EndEntry.Step"))
        assert first_step_goto == len(test_steps)
        assert first_step_goto not in [1, 2]

        converted_fp, rule_fps = converter.convert(path, TEST_FILE_DIR, "to_delete")
        os.remove(converted_fp)
        for rule_fp in rule_fps:
            os.remove(rule_fp)



step_with_bounds_template = (
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
    "<TestStep>\n"
    #  mispelling taken directly from sample file
    "  <StepType>Dischrge</StepType>\n"
    "  <StepMode>Current </StepMode>\n"
    "  <StepValue>1.0</StepValue>\n"
    "  <Limits/>\n"
    "  <Ends>\n"
    "    <EndEntry>\n"
    "      <EndType>Voltage </EndType>\n"
    "      <SpecialType> </SpecialType>\n"
    "      <Oper>&lt;= </Oper>\n"
    "      <Step>002</Step>\n"
    "      <Value>{voltage_v_lowerbound}</Value>\n"
    "    </EndEntry>\n"
    "    <EndEntry>\n"
    "      <EndType>Voltage </EndType>\n"
    "      <SpecialType> </SpecialType>\n"
    "      <Oper>&gt;= </Oper>\n"
    "      <Step>002</Step>\n"
    "      <Value>{voltage_v_upperbound}</Value>\n"
    "    </EndEntry>\n"
    "    <EndEntry>\n"
    "      <EndType>Current </EndType>\n"
    "      <SpecialType> </SpecialType>\n"
    "      <Oper>&lt;= </Oper>\n"
    "      <Step>002</Step>\n"
    "      <Value>{current_a_lowerbound}</Value>\n"
    "    </EndEntry>\n"
    "    <EndEntry>\n"
    "      <EndType>Current </EndType>\n"
    "      <SpecialType> </SpecialType>\n"
    "      <Oper>&gt;= </Oper>\n"
    "      <Step>002</Step>\n"
    "      <Value>{current_a_upperbound}</Value>\n"
    "    </EndEntry>\n"
    "  </Ends>\n"
    "  <Reports></Reports>\n"
    "  <Range>A</Range>\n"
    "  <Option1>N</Option1>\n"
    "  <Option2>N</Option2>\n"
    "  <Option3>N</Option3>\n"
    "  <StepNote></StepNote>\n"
    "</TestStep>\n"
)